import asyncio
import re
import pickle
from .ack_udp import *
from .net import *

TYPE_UDP_CON = 1
TYPE_UDP_SERVER = 2
TYPE_TCP_CON = 3
TYPE_TCP_SERVER = 4
TYPE_TCP_CLIENT = 5
SUB_ALL = [b"", b""]

"""
The code in this class supports a pull / fetch style use-case.
More suitable for some apps whereas the parent class allows
for handles to handle messages as they come in. The fetching
API needs for messages to be subscribed to beforehand.
"""
class BaseStream(ACKUDP):
    def __init__(self, proto, loop=None, conf=NET_CONF):
        super().__init__()
        self.conf = conf
        self.dest = None
        self.dest_tup = None
        self.loop = loop or asyncio.get_event_loop()

        # [Bool(msg)] = Queue.
        # Lets convert this to [b"msg pattern", b"host pattern"] = [Queue]
        self.subs = {}

        # Instance of the base proto class.
        self.proto = proto
        self.route = self.proto.route

        # Used for doing send calls.
        self.handle = {}

    """
    (1) UDP is multiplexed and doesn't need a destination bound.
    (2) TCP cons have a dest set.
    (3) TCP and UDP servers won't have a dest.
    """
    def set_dest_tup(self, dest_tup):
        self.dest_tup = dest_tup

    """
    Set internal handle used for doing sends.
    For UDP this is a asyncio.DatagramTransport.
    For TCP it's a asyncio.StreamWriter.
    """
    def set_handle(self, handle, client_tup=None):
        if client_tup is not None:
            self.handle[client_tup] = handle
        else:
            self.handle = handle

    def hash_sub(self, sub):
        return hash(sub[0]) + hash(sub[1])

    # Subscribe to a certain message and host type.
    # sub = [b_msg_pattern, b_addr_pattern]
    def subscribe(self, sub, handler=None):
        offset = self.hash_sub(sub)
        if offset not in self.subs:
            self.subs[offset] = [
                sub,
                asyncio.Queue(self.conf["max_qsize"]),
                handler
            ]

        return self

    # Remove a subscription.
    def unsubscribe(self, sub):
        offset = self.hash_sub(sub)
        if offset in self.subs:
            del self.subs[offset]

        return self
    
    def tup_to_sub(self, dest_tup):
        return [
            b"", # Any message.
            re.escape(b'%s:%d' % ( # Specific IP:port.
                to_b(dest_tup[0]), 
                dest_tup[1]
            ))
        ]

    # Adds a message to the first valid bucket.
    """
    TODO: If the queue is full what should the default
    actions be? Should the program await the queue
    being empty? Should it discard data? Maybe on limit - 1
    call https://docs.python.org/3/library/asyncio-eventloop.html#asyncio.loop.remove_reader and on queue empty add it back.
    """
    def add_msg(self, data, client_tup):
        # No subscriptions.
        if not len(self.subs):
            log("no subs")
            return

        # Add message to queue and raise an event.
        def do_add(q):
            # Check queue isn't full.
            if q.full():
                # TODO: Remove sock from event select.
                # To give time for queue to be processed.
                q.get_nowait()

            # Put an item on the queue.
            assert(isinstance(client_tup, tuple))
            q.put_nowait([client_tup, data])

        # Apply bool filters to message.
        msg_added = False
        client_addr = b"%s:%d" % (to_b(client_tup[0]), client_tup[1])
        for sub, q, handler in self.subs.values():
            # Msg pattern, address pattern.
            b_msg_p, b_addr_p = sub

            # Check client_addr matches their host pattern.
            if b_addr_p:
                host_matches = re.findall(b_addr_p, client_addr)
                if host_matches == []:
                    continue

            # Check data matches their message pattern.
            if b_msg_p:
                msg_matches = re.findall(b_msg_p, data)
                if msg_matches == []:
                    continue

            # Execute message using handle instead of adding to queue.
            if handler is not None:
                log("handler not none.")
                run_handler(self.proto, handler, client_tup, data)
                continue

            # Add message to queue.
            msg_added = True
            do_add(q)

        if not msg_added:
            log(f"Discarded {client_tup} = {data}")

    # Async wait for a message that matches a pattern in a queue.
    async def recv(self, sub=SUB_ALL, timeout=2, full=False):
        recv_timeout = timeout or self.conf["recv_timeout"]
        offset = self.hash_sub(sub)
        try:
            # Sanity checking.
            if offset not in self.subs:
                raise Exception("Sub not found. Forgot to subscribe.")

            # Get message from queue with timeout.
            _, q, handler = self.subs[offset]
            ret = await asyncio.wait_for(
                q.get(),
                recv_timeout
            )

            # Run handler if one is set.
            if handler is not None:
                run_handler(self.proto, handler, ret[0], ret[1])

            # Return data, sender_tup.
            if full:
                return ret
            else:
                # Return only the data portion.
                return ret[1]
        except Exception as e:
            return None

    # Async send for TCP and UDP cons.
    # Listen servers also supported.
    async def send(self, data, dest_tup):
        try:
            # Get handle reference.
            if isinstance(self.handle, dict):
                handle = self.handle[dest_tup]
            else:
                handle = self.handle

            # TCP send -- already bound to a target.
            # Indexed by writer streams per con.
            if isinstance(handle, asyncio.streams.StreamWriter):
                handle.write(data)
                await handle.drain()
                return 1

            # UDP send -- not connected - can be sent to anyone.
            # Single handle for multiplexing.
            if isinstance(handle, DATAGRAM_TYPES):
                handle.sendto(
                    data,
                    dest_tup
                )
                return 1

            # TCP send -- already bound to transport con.
            # TCP Transport instance.
            if isinstance(handle, STREAM_TYPES):
                # This also works for SSL wrapped sockets.
                handle.write(data)
                """
                await self.loop.sock_sendall(
                    self.proto.sock,
                    data
                )
                """
                

                return 1

            return 0
        except Exception as e:
            log(f"{self.handle}")
            log_exception()
            return 0

"""
In Python's asyncio code you can use so-called 'protocol' classes
to receive messages from an endpoint or server and then handle
them in real time. This is a very elegant way to do things
because the event loop handles polling the sockets to check
if there's any new messages for you vs you doing the check
yourself using await. The drawback is the protocol-style way
of networking basically uses callbacks and by itself -- doesn't
mix well with the async way of doing things. But with some small
tweaks it is flexible enough to do whatever you want.
"""
class BaseProto(BaseACKProto):
    def __init__(self, sock, route=None, loop=None, conf=NET_CONF):
        super().__init__(conf)

        # Config.
        self.conf = conf
        self.loop = loop

        # Socket of underlying connection.
        self.client_tup = None
        self.sock = sock
        self.tcp_clients = []
        self.tcp_server = None
        self.tcp_server_task = None
        self.endpoint_type = None

        # Used for TCP server awaitable.
        self.p_client_entry = 0 # Location of the pipe in client futures.
        self.p_client_insert = 0 # Last insert location in client futures.
        self.p_client_get = 0 # Offset that increases per await over the futures.
        self.client_futures = { 0: asyncio.Future() } # Table for TCP client pipes.

        # Bind / route.
        self.route = route

        # Can have pipes to other streams that it broadcasts to.
        self.pipes = []

        # List of other pipes that pipe to this.
        self.parent_pipes = []

        # Process messages in real time.
        self.msg_cbs = []

        # Ran when a connection ends.
        self.end_cbs = []

        # Ran when a connection is made.
        # For TCP this is a new connection.
        self.up_cbs = []

        # List of tasks for send / recv / subscribe.
        """
        Coroutine references need to be saved or the garbage collector
        may clean them up. The list here is a generic list for async
        operations in motion as part of using this class -- possible
        operations like send / recv called via msg handlers may end
        up here and they're awaited for completion.
        """
        self.tasks = []

        # Tasks saved for running msg handlers.
        """
        When any message handlers that are coroutines are registered
        and run on a new message it's saved as a task in this list.
        These tasks aren't awaited for completion in close so
        that message handlers can call close themselves and not cause
        infinite waiting loops (their handler would never be 'done.'
        as it awaits on themself to finish.)
        Cleaned-up when a handler task is done.
        """
        self.handler_tasks = []

        # For unique messages if enabled.
        self.msg_ids = {}

        # Event fired when stream set.
        self.stream_ready = asyncio.Event()

        # Placeholders.
        self.transport = None
        self.stream = None
        self.is_ack = None
        self.is_ackable = None
        self.is_running = True

    # Indicates the type of endpoint this is.
    def set_endpoint_type(self, endpoint_type):
        self.endpoint_type = endpoint_type

    # Used for event-based programming.
    # Can execute code on new cons, dropped cons, and new msgs.
    def run_handlers(self, handlers, client_tup=None, data=None):
        # Run any registered call backs on msg.
        self.handler_tasks = rm_done_tasks(self.handler_tasks)
        for handler in handlers:
            # Run the handler as a callback or coroutine.
            run_handler(self, handler, client_tup, data)

    def get_client_tup(self):
        # Get transport address.
        client_tup = None
        if self.sock is not None:
            # Use local socket details (for servers.)
            client_tup = self.sock.getsockname()
            
            # Try use remote peer info if it exists.
            try:
                client_tup = self.sock.getpeername()
            except Exception:
                pass

        return client_tup

    def add_tcp_client(self, client):
        # Save location of this client pipe in the table.
        client.p_client_entry = self.p_client_insert

        # Point to next entry in table and initialize it.
        self.p_client_insert = (self.p_client_insert + 1) % sys.maxsize
        self.client_futures[self.p_client_insert] = asyncio.Future()

        # Store this pipe in the Future.
        self.tcp_clients.append(client)
        self.client_futures[client.p_client_entry].set_result(client)

    async def make_awaitable(self):
        if self.endpoint_type == TYPE_TCP_SERVER:
            bound = self.p_client_insert + 1
            for p in range(0, bound):
                # Get reference to the current future to await on.
                cur_p_get = (self.p_client_get + p) % bound

                # Skip empty entries deleted on connection lost.
                if self.client_futures[cur_p_get] is None:
                    continue

                # Increment the pointer to the next future in line.
                # This sets it up for the next await call to work.
                # Only increment it if the current location is taken.
                if self.client_futures[cur_p_get].done():
                    self.p_client_get = (cur_p_get + 1) % bound

                # Await on the future at the head of the futures.
                return await self.client_futures[cur_p_get]

            raise Exception("Could not find awaitable future accept().")
        else:
            # TCP con -> one pipe so no reason to await it.
            # UDP server or con -> multiplex so one pipe for everything.
            return self

    def __await__(self):
        return self.make_awaitable().__await__()

    def set_tcp_server(self, server):
        self.transport = server
        self.tcp_server = server

    def set_tcp_server_task(self, task):
        self.tcp_server_task = task

    def set_ack_handlers(self, is_ack, is_ackable):
        self.is_ack = is_ack
        self.is_ackable = is_ackable
        return self

    def add_pipe(self, pipe):
        self.pipes.append(pipe)
        pipe.parent_pipes.append(self)
        return self

    def del_pipe(self, pipe):
        if pipe in self.pipes:
            self.pipes.remove(pipe)

        return self

    def add_msg_cb(self, msg_cb):
        self.msg_cbs.append(msg_cb)
        return self

    def del_msg_cb(self, msg_cb):
        if msg_cb in self.msg_cbs:
            self.msg_cbs.remove(msg_cb)

        return self
    
    def add_up_cb(self, up_cb):
        self.up_cbs.append(up_cb)
        return self

    def del_up_cb_cb(self, up_cb):
        if up_cb in self.up_cbs:
            self.up_cbs.remove(up_cb)

        return self

    def add_end_cb(self, end_cb):
        self.end_cbs.append(end_cb)

        # Make sure it runs if this is already closed.
        if not self.is_running:
            self.run_handlers(end_cb)

        return self

    def del_end_cb(self, end_cb):
        if end_cb in self.end_cbs:
            self.end_cbs.remove(end_cb)

        return self

    # Called only once for UDP.
    def connection_made(self, transport):
        if self.stream is None:
            # Record the endpoint.
            if transport is not None:
                self.transport = transport
                self.client_tup = self.get_client_tup()

            # Set stream object for doing I/O.
            self.stream = BaseStream(self, loop=self.loop)
            self.stream_ready.set()

        # Process messages using any registered handlers.
        self.run_handlers(self.up_cbs)

    # Socket closed manually or shutdown by other side.
    def connection_lost(self, exc):
        super().connection_lost(exc)

        # Remove self from any parent pipes.
        for pipe in self.parent_pipes:
            pipe.del_pipe(self)
        
        # Execute any cleanup handlers.
        self.run_handlers(self.end_cbs, self.client_tup)

    def route_msg(self, data, client_tup):
        # No data to route.
        if not data:
            return

        # Route messages to any pipes.
        for pipe in self.pipes:
            task = asyncio.create_task(
                pipe.send(
                    data,
                    pipe.stream.dest_tup
                )
            )

            self.tasks.append(task)

        # Add message to any interested subscriptions.
        # Matching pattern for host is in bytes so
        # there is a need to convert ip to bytes.
        self.stream.add_msg(
            data,
            (client_tup[0], client_tup[1])
        )

        # Process messages using any registered handlers.
        self.run_handlers(self.msg_cbs, client_tup, data)

    def handle_data(self, data, client_tup):
        # Convert data to bytes.
        if isinstance(data, bytearray):
            data = bytes(data)

        # Record msg received.
        log(
            'data recv {} = {}'.format(client_tup, data)
        )

        # Ack UDP msg if enabled.
        if self.is_ack and self.is_ackable:
            """
            Sends an ACK down the stream if it's a message that needs an ACK.
            Clients that use the 'reliable' UDP functions over a specific
            protocol provide their own functions for returning these ACKs.
            Hence the code works with any protocol.
            """
            did_ack, payload = self.stream.handle_ack(
                data,
                self.is_ack,
                self.is_ackable,
                lambda buf: self.stream.send(buf, client_tup)
            )

            """
            The Stream protocol class does not route back
            messages that are ACKs to messages we sent. Otherwise
            a sender might see a returned ACK and get into a loop
            trying to ACK it themself. It's a control message so
            there's no real reason to route them to recv.
            """
            if not did_ack:
                return
            else:
                # Strip the header portion out.
                data = payload

        # Supports unique messages.
        if self.conf["enable_msg_ids"]:
            if not self.is_unique_msg(self.stream, data, client_tup):
                log("not unique.")
                return

        # Route message to stream.
        self.route_msg(data, client_tup)

    def error_received(self, exp):
        pass

    # UDP packets.
    def datagram_received(self, data, client_tup):
        log(f"Base proto recv udp = {data}")
        if self.transport is None:
            log(f"Skipping process data cause transport none 1.")
            return

        self.handle_data(data, client_tup)

    # Single TCP connection.
    def data_received(self, data):
        log(f"Base proto recv tcp = {data}")
        if self.transport is None:
            log(f"Skipping process data cause transport none 2.")
            return

        self.handle_data(
            data,
            self.transport.get_extra_info('socket').getpeername()
        )

    async def close(self, do_sleep=True):
        # Already closed.
        if not self.is_running:
            return

        # Skip sleep for TCP clients.
        if do_sleep:
            await asyncio.sleep(0)

        """
        # Cancel serve forever.
        if self.tcp_server_task is not None:
            self.transport.shutdown()
        """

        # Wait for sending tasks in ACK UDP.
        if self.stream is not None:
            # Set ACKs for all sent messages.
            for seq_no in self.stream.seq.keys():
                self.stream.seq[seq_no].set()

            # Wait for all send loops to end.
            if len(self.stream.ack_send_tasks):
                await gather_or_cancel(self.stream.ack_send_tasks, 4)
                self.stream.ack_send_tasks = []

        # Wait for all current tasks to end.
        #self.tasks = rm_done_tasks(self.tasks)
        self.tasks = []
        if len(self.tasks):
            # Wait for tasks to finish.
            await gather_or_cancel(self.tasks, 4)
            self.tasks = []

        # Close spawned TCP clients for a TCP server.
        for client in self.tcp_clients:
            # Client is already closed.
            if not client.is_running:
                continue

            # Close client transports.
            await client.close(do_sleep=False)

        # Close the main server socket.
        # This does cleanup for any TCP servers.
        if self.transport is not None:
            self.transport.close()

        """
        if self.tcp_server_task is not None:
            self.tcp_server_task.cancel()
        """

        # Close any sockets.
        # if self.sock is not None:
        #    self.sock.close()
        
        # No longer running.
        self.transport = None
        self.socket = None
        self.is_running = False
        self.tcp_server = None
        self.tcp_server_task = None
        self.tcp_clients = []

    # Return a matching message, async, non-blocking.
    async def recv(self, sub=SUB_ALL, timeout=2, full=False):
        return await self.stream.recv(sub, timeout, full)

    async def send(self, data, dest_tup=None):
        dest_tup = dest_tup or self.stream.dest_tup
        return await self.stream.send(data, dest_tup)

    # Sync subscribe to a message.
    # Easy way to get a message from sync code too.
    def subscribe(self, sub=SUB_ALL, handler=None):
        self.stream.subscribe(sub, handler)
        return self

    def unsubscribe(self, sub):
        self.stream.unsubscribe(sub)
        return self

    # Echo client just for testing.
    async def echo(self, msg, dest_tup):
        buf = bytearray().join([b"ECHO ", msg])
        await self.send(buf, dest_tup)

"""
StreamReaderProtocol provides a way to "translate" between
Protocol and StreamReader. Mostly we're interested in having
a protocol class for TCP that can handle messages as they're
ready as opposed to having to poll ourself. Encapsulates
a client connection to a TCP server in a BaseProto object.
"""
class BaseStreamReaderProto(asyncio.StreamReaderProtocol):
    def __init__(self, stream_reader, base_proto, loop, conf=NET_CONF):
        # Setup stream reader / writers.
        super().__init__(stream_reader, lambda x, y: 1, loop=loop)

        # This is the server that spawns these client connections.
        self.proto = base_proto
        self.loop = loop

        # Will represent us.
        # Servers route above will be reused for this.
        self.client_proto = None
        self.client_offset = 0

        # Main class variables.
        self.sock = None
        self.transport = None
        self.remote_tup = None
        self.conf = conf

    """
    StreamReaderProtocol has a bug in this function and doesn't
    properly return False. This is a patch.
    """
    def eof_received(self):
        #self.transport.pause_reading()
        reader = self._stream_reader
        if reader is not None:
            reader.feed_eof()
            
        return False

    def connection_made(self, transport):
        # Wrap this connection in a BaseProto object.
        self.transport = transport
        self.sock = transport.get_extra_info('socket')
        self.remote_tup = self.sock.getpeername()
        self.client_proto = BaseProto(
            sock=self.sock,
            route=self.proto.route,
            conf=self.conf,
            loop=self.loop
        )

        # Log connection details.
        log(f"New TCP client l={self.sock.getsockname()}, r={self.remote_tup}")

        # Setup stream object.
        self.client_proto.set_endpoint_type(TYPE_TCP_CLIENT)
        self.client_proto.msg_cbs = self.proto.msg_cbs
        self.client_proto.end_cbs = self.proto.end_cbs
        self.client_proto.up_cbs = self.proto.up_cbs
        self.client_proto.connection_made(transport)

        # Record destination.
        self.client_proto.stream.set_dest_tup(self.remote_tup)

        # Record instance to allow cleanup in server.
        self.proto.add_tcp_client(self.client_proto)

        # Setup handle for writing.
        super().connection_made(transport)
        self.client_proto.stream.set_handle(
            self._stream_writer,

            # Index writers by peer connection.
            self.remote_tup
        )


    
    # If close was called on a pipe on a server then clients will already be closed.
    # So this code will have no effect.
    def connection_lost(self, exc):
        super().connection_lost(exc)

        # Cleanup client futures entry.
        p_client_entry = self.client_proto.p_client_entry
        client_future = self.proto.client_futures[p_client_entry]
        if client_future.done():
            del self.proto.client_futures[p_client_entry]

        # Run disconnect handlers if any set.
        client_tup = self.remote_tup
        self.client_proto.run_handlers(self.client_proto.end_cbs, client_tup)

        # Close its client socket and transport.
        try:
            if self.client_proto in self.proto.tcp_clients:
                self.proto.tcp_clients.remove(self.client_proto)
        except Exception:
            log_exception()

        try:
            self.transport.close()
            self.transport = None
        except Exception:
            log_exception()

        # Remove this as an object to close and manage in the server.
        super().connection_lost(exc)

    def error_received(self, exp):
        pass

    def data_received(self, data):
        # This just adds data to reader which we are handling ourselves.
        #super().connection_lost(exc)
        if self.client_proto is None:
            return

        if not len(self.client_proto.msg_cbs):
            log("No msg cbs registered for inbound message in hacked tcp server.")

        self.client_proto.handle_data(data, self.remote_tup)

# Returns a hacked TCP server object
async def base_start_server(sock, base_proto, *, loop=None, conf=NET_CONF, **kwds):
    # Main vars.
    loop = loop or asyncio.get_event_loop()
    def factory():
        reader = asyncio.StreamReader(limit=conf["reader_limit"], loop=loop)
        return BaseStreamReaderProto(
            reader,
            base_proto,
            loop,
            conf
        )

    # Call the regular create server func with custom protocol factory.
    server = await loop.create_server(
        factory,
        sock=sock,
        **kwds
    )

    return server

# Started in a new process.
def start_server_threaded(args):
    # Create new event loop and run coroutine in it.
    asyncio.set_event_loop_policy(SelectorEventPolicy())
    loop = asyncio.new_event_loop()
    #loop = asyncio.get_event_loop()
    f = asyncio.ensure_future(
        async_wrap_errors(
            args[0].serve_forever()
        ),
        loop=loop
    )

    loop.run_until_complete(f)
    loop.stop()

"""
In the spirit of unix a 'pipe' is an protocol and destination
agnostic way to send data. It supports TCP & UDP: cons & servers.
It supports using IPv4 and IPv6 destination addresses.
You can pull data from it based on a regex pattern.
You can execute code on new messages or connection disconnects.
"""
async def pipe_open(proto, route, dest=None, sock=None, msg_cb=None, up_cb=None, conf=NET_CONF):
    # If no route is set assume default interface route 0.
    if route is None:
        # Load internal addresses.
        i = await Interface().start_local()

        # Bind to route 0.
        route = await i.route()

    # If dest has no route set use this route.
    if dest is not None and dest.route is None:
        if not dest.resolved:
            dest.route = route
            await dest

    # Build the base protocol object.
    base_proto = None
    try:
        # Get event loop reference.
        if conf["loop"] is not None:
            loop = conf["loop"]()
        else:
            loop = asyncio.get_event_loop()

        # Build socket bound to specific interface.
        if sock is None:
            sock = await socket_factory(
                route=route,
                dest_addr=dest,
                sock_type=UDP if proto == RUDP else proto,
                conf=conf
            )

            # Check if sock succeeded.
            if sock is None:
                log("Could not allocate socket.")
                return None

            # Connect socket if TCP.
            if proto == TCP and dest is not None:
                # Set non-blocking.
                sock.settimeout(0)
                sock.setblocking(0)

                # Connect the socket task.
                con_task = asyncio.create_task(
                    loop.sock_connect(
                        sock, 
                        dest.tup
                    )
                )
                
                # Wait for connection, async style.
                await asyncio.wait_for(con_task, conf["con_timeout"])

                # Enable SSL on this socket.
                """
                if conf["use_ssl"]:
                    
                    # Some security options are disabled for simplicity.
                    # TODO: explore this more.
                    ssl_context = ssl.create_default_context()
                    ssl_context.check_hostname = False 
                    ssl_context.verify_mode = ssl.CERT_NONE
                    

                    # Wrap socket won't support non-blocking sockets.
                    # Temporarily make it blocking.
                    sock.settimeout(conf["ssl_handshake"])

                    # The socket is wrapped in an SSL context after all
                    # socket options are set.
                    sock = ssl_context.wrap_socket(
                        sock,

                        # Hostname validation looks like this.
                        # But will only work if the dest isn't an IP.
                        #server_hostname=dest.tup[1]
                    )

                    # Then the socket is made non-blocking again.
                    sock.settimeout(0)
                """
                    
        # Make sure bind port is set (and not zero.)
        route.bind_port = sock.getsockname()[1]

        # Return the sock instead of base proto.
        if conf["sock_only"]:
            return sock

        # Main protocol instance for routing messages.
        #if base_proto is None:
        base_proto = BaseProto(sock=sock, route=route, loop=loop, conf=conf)

        # Add message handler.
        if msg_cb is not None:
            base_proto.add_msg_cb(msg_cb)

        # Start processing messages for UDP.
        if proto in [UDP, RUDP]:
            transport, _ = await loop.create_datagram_endpoint(
                lambda: base_proto,
                sock=sock
            )

            await base_proto.stream_ready.wait()
            base_proto.stream.set_handle(transport, client_tup=None)
            if dest is not None:
                base_proto.set_endpoint_type(TYPE_UDP_CON)
            else:
                base_proto.set_endpoint_type(TYPE_UDP_SERVER)

        # Install default ack builder and handler.
        # Now it is poorman's TCP ;_____; but still no ordering.
        if proto == RUDP:
            base_proto.set_ack_handlers(
                is_ack=base_proto.stream.is_ack,
                is_ackable=base_proto.stream.is_ackable
            )

        # Start processing messages for TCP.
        if proto == TCP:
            # Add new connection handler.
            if up_cb is not None:
                base_proto.add_up_cb(up_cb)

            # Listen server.
            if dest is None:
                # Start router for TCP messages.
                server = await base_start_server(
                    sock=sock,
                    base_proto=base_proto,
                    loop=loop,
                    conf=conf
                )

                # Make the server start serving requests.
                assert(server is not None)
                base_proto.set_tcp_server(server)

                # Saving the task is apparently needed
                # or the garbage collector could close it.
                if hasattr(server, "serve_forever"):
                    server_task = asyncio.create_task(
                        async_wrap_errors(
                            server.serve_forever()
                        )
                    )
                    
                    base_proto.set_tcp_server_task(server_task)
                    #asyncio.ensure_future(server_task)

                    """
                    threading.Thread(target=server_task.serve_forever).start()
                    #await server_task
                    
                    loop.run_in_executor(
                        None, start_server_threaded, (server,)
                    )
                    """

                base_proto.set_endpoint_type(TYPE_TCP_SERVER)

            # Single connection.
            if dest is not None:
                # Enable SSL on this socket.
                if conf["use_ssl"]:
                    # Some security options are disabled for simplicity.
                    # TODO: explore this more.
                    ssl_context = ssl.create_default_context()
                    ssl_context.check_hostname = False 
                    ssl_context.verify_mode = ssl.CERT_NONE
                    server_hostname = ""
                else:
                    ssl_context = False
                    server_hostname = None

                # base_proto.set_handle(writer, sock.getpeername())
                await loop.create_connection(
                    protocol_factory=lambda: base_proto,
                    sock=sock,
                    ssl=ssl_context,
                    server_hostname=server_hostname
                )

                # Set transport handle.
                await base_proto.stream_ready.wait()
                base_proto.stream.set_handle(base_proto.transport, dest.tup)
                base_proto.set_endpoint_type(TYPE_TCP_CON)

        # Set dest if it's present.
        if dest is not None:
            base_proto.stream.dest = dest
            base_proto.stream.set_dest_tup(dest.tup)

            # Queue all messages for convenience.
            base_proto.subscribe(SUB_ALL)

        # Register pipes, msg callbacks, and subscriptions.
        return base_proto
    except Exception as e:
        log_exception()
        if conf["no_close"] == False:
            log("no close is false so trying to clean up.")
            try:
                if sock is not None:
                    log(f"closing socket. {sock.getsockname()}")
                    sock.close()
                

                if base_proto is not None:
                    log("closing bas proto")
                    await base_proto.close()
            except:
                log_exception()

if __name__ == "__main__": # pragma: no cover
    from .interface import Interface
    from .http_client import http_req
    from .address import Address
    async def test_base():
        af = IP4; proto = TCP; port = 10101
        interface = await Interface("enp3s0").start()
        route = await interface.route(af).bind(port)

        """
        TCP server works.
        # A transport router for TCP.
        # Spawms a new stream reader protocol per con.
        base_proto = await transport_router(
            route=route,
            proto=proto
        )
        """

        dest_addr = await Address("127.0.0.1", 12344, route).res()
        base_proto = await pipe_open(
            route=route,
            proto=proto,
            dest=dest_addr
        )

        # Simple message callback that prints inbound data and client addr tups.
        base_proto.add_msg_cb(
            # Data, client tup, stream, route
            lambda d, c, s: print(d, c)
        )

        # Build HTTP req and send to Google con writer.
        get_req = http_req(dest_addr)
        await base_proto.stream.send(get_req, dest_addr.tup)

        # Don't close the main coroutine.
        while 1:
            await asyncio.sleep(1)

    async_test(test_base)


import copy
from http.client import HTTPResponse
import json
from .net import *
from .base_stream import *
from .address import *

HTTP_HEADERS = [
    [b"User-Agent", b"curl/7.54.0"],
    [b"Origin", b"null"],
    [b"Accept", b"*/*"]
]

def http_req_buf(af, host, path=b"/", method=b"GET", payload=b"", headers=None):
    # Format headers.
    hdrs = {}
    if headers is None:
        headers = HTTP_HEADERS
    else:
        headers += HTTP_HEADERS

    # Raw http request.
    # Very important: 1.0 is used to disable 'chunked encoding.'
    # Chunked encoding overly complicates processing HTTP responses.
    buf  = b"%s %s HTTP/1.0\r\n" % (to_b(method), to_b(path))
    if af == IP4:
        host = to_b(host)
    else:
        host = to_b(f"[{to_s(host)}]")
    buf += b"Host: %s\r\n" % (host)
    for header in headers:
        n, v = header

        # Don't add host header twice.
        if n.lower() == b"host":
            continue
        
        # Skip duplicate headers.
        if n not in hdrs:
            buf += b"%s: %s\r\n" % (n, v)
            hdrs[n] = 1

    # Add content length for payload.
    if payload is not None:
        buf += to_b(f"Content-Length: {len(payload)}\r\n")
    
    # Terminate headers.
    buf = buf[:-2]
    assert(buf[-1] not in [b'\r', b'\n'])
    buf += b"\r\n\r\n"

    # Append payload (if any.)
    if payload is not None:
        buf += to_b(payload)

    return buf

def http_parse_headers(self):
    # Get headers from named pair list.
    hdrs = {}
    for named_pair in self.headers._headers:
        name, value = named_pair
        hdrs[name] = value
        hdrs[name.lower()] = value

    # Set origin.
    if 'Origin' not in hdrs:
        hdrs['Origin'] = 'null'
        hdrs['origin'] = 'null'

    # Save header list.
    self.hdrs = hdrs

class ParseHTTPResponse(HTTPResponse):
    def __init__(self, resp_text):
        self.resp_len = len(resp_text)
        self.sock = FakeSocket(resp_text)
        super().__init__(self.sock)
        self.begin()
        http_parse_headers(self)

        te = "Transfer-Encoding"
        if te in self.hdrs:
            if self.hdrs[te] == "chunked":
                raise Exception("chunked encodign not supported!")

    def out(self):
        return self.read(self.resp_len)
    
def get_hdr(name, hdrs):
    # Hdrs none probably.
    if not isinstance(hdrs, list):
        return (-1, None)
    
    # Look for particular HTTP header.
    for index, hdr in enumerate(hdrs):
        if hdr[0].lower() == name.lower():
            return (index, hdr[1])
        
    # Not found.
    return (-1, None)

async def http_req(route, dest, path, do_close=1, method=b"GET", payload=None, headers=None, conf=NET_CONF):
    # Get a new con 
    r = copy.deepcopy(route)
    r = await r.bind()
    
    assert(dest is not None)
    log(f"{route} {dest}")
    try:
        p = await pipe_open(route=r, proto=TCP, dest=dest, conf=conf)
    except Exception:
        log_exception()

    if p is None:
        return None, None

    try:
        p.subscribe(SUB_ALL)

        # Set host and port from con service.
        host, port = dest.tup

        # But overwrite host if it's set.
        hdr_index, new_host = get_hdr(b"Host", headers)
        if new_host is not None:
            host = new_host
            del headers[hdr_index]

        # Build raw HTTP request.
        buf = http_req_buf(
            route.af,
            host,
            path,
            method=method,
            payload=payload,
            headers=headers
        )

        await p.send(buf, dest.tup)
        out = await p.recv(SUB_ALL)
    except Exception:
        log_exception()
        await p.close()
        return None, None

    if do_close:
        await p.close()
        p = None

    if out is not None:
        out = ParseHTTPResponse(out)
        
    return p, out

"""
Break up a URL into it's host, port, file path,
and resolve it's domain as an address for use with
networking code that works to make the HTTP request.
"""
async def url_res(route, url, timeout=3):
    # Split URL into host and port.
    port = 80
    url_parts = urllib.parse.urlparse(url)
    host = netloc = url_parts.netloc
    path = url_parts.path

    # Overwrite default port 80.
    if ":" in netloc:
        host, port = netloc.split(":")
        port = int(port)

    # Resolve domain of URL.
    dest = await Address(
        host,
        port,
        route,
        timeout=timeout
    )

    # Return URL parts.
    return {
        "host": host,
        "port": port,
        "path": path,
        "dest": dest
    }

"""
The purpose of this function is to provide something simple for
fetching a remote URL. The URL param may be pre-resolved using url_res,
otherwise it can be a URL string to fetch. The function accepts optional
GET params. The post method is not supported.
"""
async def url_open(route, url, params={}, timeout=3, throttle=0, headers=[], conf=NET_CONF):
    # Other types for url are not supported.
    url_parts = Exception("url param type is not supported.")

    # url is a URL so break into parts.
    if isinstance(url, str):
        url_parts = await url_res(route, url, timeout=timeout)

    # URL is provided as parts already.
    if isinstance(url, dict):
        url_parts = url

    # Throttle request.
    if throttle:
        await asyncio.sleep(throttle)

    # Encode GET params.
    get_vars = ""
    for name in params:
        # Encode the value.
        v = to_s(
            urlencode(
                params[name]
            )
        )

        # Pass the params in the URL.
        n = to_s(urlencode(name))
        get_vars += f"&{n}={v}"

    # Request path.
    if len(get_vars):
        path = f"{url_parts['path']}?{get_vars}"
    else:
        path = url_parts["path"]

    # Make req.
    conf["con_timeout"] = timeout
    conf["recv_timeout"] = timeout
    headers += [[b"Host", to_b(url_parts["host"])]]
    _, resp = await http_req(
        route,
        url_parts["dest"],
        path,

        # Specify the right sub/domain.
        headers=headers,
        conf=conf
    )

    # Return API output as string.
    return to_s(resp.out())

# Web payload decorators
def Payload(f, url={}, body=b""):
    async def wrapper(path, hdrs=[]):
        return await f(path=path, hdrs=hdrs, url=url, body=body)

    return wrapper

# urllib.parse.urlencode(params)

# Returns pipe, ParseHTTPResponse
async def do_web_req(addr, http_buf, do_close, conf=NET_CONF):
    log(f"{addr.route} {addr}")

    # Open TCP connection to HTTP server.
    route = addr.route.interface.route()
    route = await route.bind()
    try:
        p = await pipe_open(
            route=route,
            proto=TCP,
            dest=addr,
            conf=conf
        )
    except Exception:
        log_exception()

    # Error return empty.
    if p is None:
        return None, None

    # Send HTTP request.
    try:
        p.subscribe(SUB_ALL)
        await p.send(http_buf, addr.tup)
        out = await p.recv(SUB_ALL, timeout=conf['recv_timeout'])
    except Exception as e:
        log_exception()
        await p.close()
        return None, None

    # Some connections may be left open.
    if do_close:
        await p.close()
        p = None

    # Parse HTTP response.
    if out is not None:
        out = ParseHTTPResponse(out)
        
    return p, out

"""
i = await Interface()
addr = await Address("www.example.com", 80, i.route())
curl = WebCurl(addr, do_close=0)
resp = await curl.vars(url_param, body_payload).get("/")
resp.pipe # http con if open
resp.out # http reply
resp.info # parsed http reply
"""

class WebCurl():
    def __init__(self, addr, throttle=0, do_close=1, hdrs=[]):
        # Address addr.route
        self.addr = addr
        self.url_params = {}
        self.hdrs = hdrs
        self.body = b""
        self.req_buf = self.out = None
        self.path = self.info = None
        self.throttle = throttle
        self.do_close = do_close

    # Figure out less brainlet way to do this.
    def copy(self):
        client = WebCurl(self.addr)
        client.url_params = self.url_params
        client.body = self.body
        client.path = self.path
        client.hdrs = self.hdrs
        client.info = self.info
        client.out = self.out
        client.req_buf = self.req_buf
        client.throttle = self.throttle
        client.do_client = self.do_close
        return client

    def vars(self, url_params={}, body=b""):
        # Avoid race conditions.
        client = self.copy()

        # Url encode url params if set.
        if len(url_params):
            client.url_params = {
                "safe": urllib.parse.urlencode(url_params),
                "unsafe": url_params
            }

        client.body_payload = body
        return client
    
    async def api(self, method, path, hdrs, conf):
        # New instance to avoid race conditions.
        client = self.copy()
        client.path = path
        client.hdrs = hdrs

        # Append url encoded path if present.
        if len(self.url_params):
            path += f'?{self.url_params["safe"]}'

        # If payload is a dict convert to json buf.
        hdrs = hdrs or self.hdrs
        if isinstance(self.body_payload, dict):
            self.body_payload = json.dumps(self.body_payload)
            hdrs.append([b"Content-Type", b"application/json"])

        # Build a HTTP request to send to server.
        req_buf = http_req_buf(
            af=self.addr.af,
            host=self.addr.host,
            path=path,
            method=method,
            payload=self.body_payload,
            headers=hdrs
        )

        # Save request for debugging.
        if IS_DEBUG:
            client.req_buf = req_buf

        # Throttle request.
        if self.throttle:
            await asyncio.sleep(self.throttle)

        # Make the HTTP request to the server.
        pipe, info = await do_web_req(
            addr=self.addr, 
            http_buf=req_buf,
            do_close=self.do_close,
            conf=conf
        )

        # Save output.
        client.pipe = pipe
        if info is None:
            client.info = None
            client.out = b""
        else:
            client.out = info.out()
            client.info = info

        return client

    async def get(self, path, hdrs=[], conf=NET_CONF):
        return await self.api("GET", path, hdrs, conf)

    async def post(self, path, hdrs=[], conf=NET_CONF):
        return await self.api("POST", path, hdrs, conf)

    async def delete(self, path, hdrs=[], conf=NET_CONF):
        return await self.api("DELETE", path, hdrs, conf)


import asyncio
import json
import re
from .test_init import *

# https://github.com/Shopify/toxiproxy/tree/main#toxics
class ToxiToxic():
    def __init__(self, toxicity=1.0, name=None):
        self.toxicity = toxicity
        self.direction = None
        self.name = name or to_s(rand_plain(20))
        self.body = None

    def copy(self):
        toxic = ToxiToxic()
        toxic.direction = self.direction
        toxic.toxicity = self.toxicity
        toxic.name = self.name
        toxic.body = self.body
        return toxic

    def downstream(self):
        toxic = self.copy()
        toxic.direction = "downstream"
        return toxic
    
    def upstream(self):
        toxic = self.copy()
        toxic.direction = "upstream"
        return toxic
    
    def api(self, j):
        j["name"] = self.name
        j["stream"] = self.direction
        j["toxicity"] = self.toxicity
        return j
    
    # Add a delay to all data going through the proxy.
    # The delay is equal to latency +/- jitter.
    def add_latency(self, ms=1000, jitter=0):
        toxic = self.copy()
        toxic.body = self.api({
            "type": "latency",
            "attributes": {
                "latency": ms,
                "jitter": jitter
            }
        })

        return toxic
    
    # Limit a connection to a maximum number of kilobytes per second.
    def add_bandwidth_limit(self, KBs=100):
        toxic = self.copy()
        toxic.body = self.api({
            "type": "bandwidth",
            "attributes": {
                "rate": KBs,
            }
        })

        return toxic

    """ 
    Delay associated dest_pipe from being closed until a timeout
    eg if its downstream it will be similar to keep alive.
    If its upstream then other toxics wont be able to close it
    until that time frame.

    EOL or graceful shutdowns are ignored (I think).

    Time ------------------------------------>
    close blocked                       close
    """
    def add_slow_close(self, ms=2000):
        toxic = self.copy()
        toxic.body = self.api({
            "type": "slow_close",
            "attributes": {
                "delay": ms,
            }
        })

        return toxic
    
    """
    Doesn't forward data, and closes the connection after timeout (close())
    If timeout is 0, the connection won't close, and data will be delayed
    until the toxic is removed.

    Time ---------------------------->
    data blocked       Optional close
    """
    def add_timeout(self, ms=2000):
        toxic = self.copy()
        toxic.body = self.api({
            "type": "timeout",
            "attributes": {
                "timeout": ms,
            }
        })

        return toxic
    
    """
    Simulate TCP RESET (Connection reset by peer) on the connections
    by closing the stub Input immediately or after a timeout.

    The behavior of Close is set to discard any unsent/unacknowledged data by setting SetLinger to 0, ~= sets TCP RST flag and resets the connection.
    Drop data since it will initiate a graceful close with FIN/ACK. (io.EOF)

    Time ----------------------->
    data sent               close (DISCARD unset and unacked data!)
    """
    def add_reset_peer(self, ms=2000):
        toxic = self.copy()
        toxic.body = self.api({
            "type": "reset",
            "attributes": {
                "timeout": ms,
            }
        })

        return toxic
    
    """
    Closes connection when transmitted data exceeded limit.

    Limit counter ---------------->
    data                     close
    """
    def add_limit_data(self, n=100):
        toxic = self.copy()
        toxic.body = self.api({
            "type": "limit_data",
            "attributes": {
                "bytes": n,
            }
        })

        return toxic

    """
    Slices TCP data up into small bits, optionally
    adding a delay between each sliced "packet".

    The SlicerToxic slices data into multiple smaller packets
    to simulate real-world TCP behavior.

    Note: the docs say that this slices them into 'bits'
    but I think this is meant informally as in pieces
    rather than a unit of measurement. The pieces it means
    are really bytes (and can only mean bytes.)

    Slicer is designed so that receipt of data needs multiple
    recv calls to get the full response. E.g. it splits up
    responses into multiple packets and sends them individually.
    This is useful for testing assumptions behind protocol clients.
    """
    def add_slicer(self, n=100, v=50, ug=20):
        toxic = self.copy()
        toxic.body = self.api({
            "type": "slicer",
            "attributes": {
                #  size in bytes of an average packet
                "average_size": n,

                # variation in bytes of an average packet < n
                "size_variation": v,

                # time in microseconds to delay each packet by
                "delay": ug
            }
        })

        return toxic

class ToxiTunnel():
    def __init__(self, name, port, client):
        self.name = name
        self.port = port
        self.client = client
        self.toxics = []

    async def new_toxic(self, toxic):
        path = f"/proxies/{self.name}/toxics"
        resp = await self.client.curl.vars(body=toxic.body).post(path)
        assert(b"error" not in resp.out)
        self.toxics.append(toxic)

    async def remove_toxic(self, toxic):
        path = f"/proxies/{self.name}/toxics/{toxic.name}"
        resp = await self.client.curl.vars().delete(path)
        assert(resp.out == b'')
        self.toxics.remove(toxic)

    async def test_list(self):
        path = f"/proxies/{self.name}"
        resp = await self.client.curl.vars().get(path)
        return resp.out

    async def get_pipe(self, conf=None):
        # Build new route.
        conf = conf or self.client.conf
        route = self.client.addr.route.interface.route()
        await route.bind()

        # Connect to the listen server for this tunnel.
        dest = await Address("localhost", self.port, route)
        pipe = await pipe_open(TCP, route, dest, conf=conf)
        return pipe, dest.tup
    
    async def get_curl(self):
        # Build new route.
        route = self.client.addr.route.interface.route()
        await route.bind()

        # Connect to the listen server for this tunnel.
        dest = await Address("localhost", self.port, route)
        return WebCurl(addr=dest, do_close=0)

    # Close the tunnel on the toxiproxy instance.
    async def close(self):
        json_body = {
            "enabled": False
        }

        path = f"/proxies/{self.name}"
        resp = await self.client.curl.vars(body=json_body).post(path)
        self.client.tunnels.remove(self)
        return resp.out

class ToxiClient():
    def __init__(self, addr, conf=NET_CONF):
        self.addr = addr
        self.conf = conf
        self.tunnels = []

    async def start(self):
        hdrs = [[b"user-agent", b"toxiproxy-cli"]]
        self.curl = WebCurl(self.addr, hdrs=hdrs)
        return self

    async def version(self):
        resp = await self.curl.vars().get("/version")
        return resp.out

    async def new_tunnel(self, addr, name=None):
        name = name or to_s(rand_plain(10))
        json_body = {
            "name": name,
            "listen": "127.0.0.1:0",
            "upstream": f"{addr.target()}:{addr.port}",
            "enabled": True
        }

        resp = await self.curl.vars(body=json_body).post("/proxies")
        assert(b"error" not in resp.out)

        # Listen porn for the tunnel server.
        port = re.findall("127[.]0[.]0[.]1[:]([0-9]+)", to_s(resp.out))[0]
        port = to_n(port)

        # Create a new object to interact with the tunnel.
        tunnel = ToxiTunnel(name=name, port=port, client=self)
        self.tunnels.append(tunnel)

        # Return tunnel.
        return tunnel

asyncio.set_event_loop_policy(SelectorEventPolicy())
async def test_setup(netifaces=None, client=None):
    # Suppress socket unclosed warnings.
    if not hasattr(test_setup, "netifaces"):
        if netifaces is None:
            test_setup.netifaces = await init_p2pd()
        else:
            test_setup.netifaces = netifaces
        #warnings.filterwarnings('ignore', message="unclosed", category=ResourceWarning)

    if not hasattr(test_setup, "client"):
        i = await Interface().start_local(skip_resolve=True)
        r = i.route()
        addr = await Address("localhost", 8474, r)
        client = ToxiClient(addr)
        await client.start()

    return client

"""
class TestToxi(unittest.IsolatedAsyncioTestCase):
    async def test_toxi_client_usage(self):
        client = await test_setup()

        # Create a new tunnel from toxiproxi to google.
        dest = await Address("www.google.com", 80, client.addr.route)
        tunnel = await client.new_tunnel(dest)
        assert(isinstance(tunnel, ToxiTunnel))

        # Test that adding and removing all toxics works.
        downstream = ToxiToxic().downstream()
        toxics = [
            downstream.add_latency(),
            downstream.add_bandwidth_limit(),
            downstream.add_slow_close(),
            downstream.add_timeout(),

            # Not in the windows binary -- newer toxic.
            #downstream.add_reset_peer(),
            downstream.add_limit_data(),
            downstream.add_slicer()
        ]

        # Try each toxic one by one.
        for toxic in toxics:
            print(f"adding {toxic.body['type']}")

            # Add the new toxic to the tunnel.
            await tunnel.new_toxic(toxic)
            assert(toxic in tunnel.toxics)

            # Check that there's a toxic for that tunnel.
            out = await tunnel.test_list()
            out = json.loads(to_s(out))
            assert(len(out["toxics"]))

            # Remove the toxic from the tunnel.
            await tunnel.remove_toxic(toxic)
            assert(toxic not in tunnel.toxics)

            # Check that there's no toxics for that tunnel.
            out = await tunnel.test_list()
            out = json.loads(to_s(out))
            assert(not len(out["toxics"]))

        # Will throw on error response.
        await tunnel.close()
"""

if __name__ == '__main__':
    pass
    #main()



import asyncio
from .p2p_node import *
from .p2p_utils import *
from .var_names import *
from .http_server_lib import *

asyncio.set_event_loop_policy(SelectorEventPolicy())

def con_info(self, p, con):
    return {
        "error": 0,
        "name": p["con_name"],
        "strategy": con.strat,
        "fd": con.sock.fileno(),
        "laddr": con.sock.getsockname(),
        "raddr": con.sock.getpeername(),
        "route": con.route.to_dict(),
        "if": {
            "name": con.route.interface.name,
            "offset": self.interfaces.index(
                con.route.interface
            )
        }
    }

class P2PDServer(Daemon):
    def __init__(self, if_list, node):
        super().__init__(if_list)
        self.node = node
        self.cons = {}

    async def msg_cb(self, msg, client_tup, pipe):
        # Parse HTTP message and handle CORS.
        req = await rest_service(msg, client_tup, pipe)

        # Output JSON response for the API.
        async def get_response():
            # Show version information.
            if req.api("/version"):
                return {
                    "title": "P2PD",
                    "author": "Matthew@Roberts.PM", 
                    "version": "0.1.0",
                    "error": 0
                }

            # All interface details.
            if req.api("(/ifs)"):
                return {
                    "ifs": if_list_to_dict(self.interfaces),
                    "error": 0
                }

            # Node's own 'p2p address.'
            if req.api("/p2p/addr"):
                return {
                    "addr": to_s(self.node.addr_bytes),
                    "error": 0
                }

            # Create a new connection and name it.
            named = ["con_name", "dest_addr"]
            p = req.api("/p2p/open/([^/]*)/([^/]*)", named)
            if p:
                # Need a unique name per con.
                if p["con_name"] in self.cons:
                    return {
                        "msg": "Con name already exists.",
                        "error": 2
                    }

                # Connect to ourself for tests.
                if p["dest_addr"] == "self":
                    p["dest_addr"] = self.node.addr_bytes

                # Attempt to make the connection.
                con, strat = await self.node.connect(
                    to_b(p["dest_addr"]),

                    # All connection strats except TURN by default.
                    P2P_STRATEGIES
                )

                # Success -- store pipe.
                if con is not None:
                    # Subscribe to any message.
                    con.subscribe(SUB_ALL)

                    # Remove con from table.
                    def build_do_cleanup():
                        def do_cleanup(msg, client_tup, pipe):
                            del self.cons[ p["con_name"] ]
                        
                        return do_cleanup

                    # Add cleanup handler.
                    con.add_end_cb(build_do_cleanup())

                    # Return the results.
                    con.strat = TXT["p2p_strat"][strat]
                    self.cons[ p["con_name"] ] = con
                    return con_info(self, p, con)

                # Failed to connect.
                if con is None:
                    return {
                        "msg": f"Con {p['con_name']} failed connect.",
                        "error": 3
                    }

            # Return con info.
            p = req.api("/p2p/con/([^/]*)", ["con_name"])
            if p:
                # Check con exists.
                if p["con_name"] not in self.cons:
                    return {
                        "msg": f"Con name {p['con_name']} not found.",
                        "error": 4,
                    }

                con = self.cons[ p["con_name"] ]
                return con_info(self, p, con)

            # Subscribe to certain message patterns.
            named = ["con_name"]
            params = ["msg_p", "addr_p"]
            defaults = [b"", b""]
            p = req.api("/p2p/sub/([^/]*)", named, params)
            if p:
                # Check con exists.
                if p["con_name"] not in self.cons:
                    return {
                        "msg": f"Con name {p['con_name']} not found.",
                        "error": 4,
                    }

                # Set default values for $_GET.
                set_defaults(p, params, defaults)
                sub = [ to_b(p["msg_p"]), to_b(p["addr_p"]) ]
                con = self.cons[ p["con_name" ] ]

                # Subscribe.
                if req.command == "GET":
                    con.subscribe(sub)

                    # Return results.
                    return {
                        "name": p["con_name"],
                        "sub": f"{sub}",
                        "error": 0
                    }

                # Unsubscribe.
                if req.command == "DELETE":
                    con.unsubscribe(sub)

                    # Return results.
                    return {
                        "name": p["con_name"],
                        "unsub": f"{sub}",
                        "error": 0
                    }
            
            # Send a text-based message to a named con.
            named = ["con_name", "txt"]
            p = req.api("/p2p/send/([^/]*)/([\s\S]+)", named)
            if p:
                # Check con exists.
                if p["con_name"] not in self.cons:
                    return {
                        "msg": f"Con name {p['con_name']} not found.",
                        "error": 4,
                    }

                # Connection to send to.
                con = self.cons[ p["con_name"] ]

                # Send data.
                await con.send(
                    data=to_b(p["txt"]),
                    dest_tup=con.stream.dest_tup
                )

                # Return success.
                return {
                    "name": p["con_name"],
                    "sent": len(p["txt"]),
                    "error": 0
                }

            # Send a text-based message to a named con.
            optional = ["timeout", "msg_p", "addr_p"]
            defaults = [0, b"", b""]
            named = ["con_name"]
            p = req.api("/p2p/recv/([^/]*)", named, optional)
            if p:
                # Check con exists.
                if p["con_name"] not in self.cons:
                    return {
                        "msg": f"Con name {p['con_name']} not found.",
                        "error": 4,
                    }

                # Set default values for $_GET.
                set_defaults(p, optional, defaults)

                # Get something from recv buffer.
                con = self.cons[ p["con_name"] ]
                try:
                    sub = [ to_b(p["msg_p"]), to_b(p["addr_p"]) ]
                    timeout = to_n(p["timeout"])
                    out = await con.recv(sub, timeout=timeout, full=True)
                    if out is None:
                        return {
                            "msg": f"recv buffer {sub} empty.",
                            "error": 6
                        }

                    return {
                        "client_tup": out[0],
                        "data": to_s(out[1]),
                        "error": 0
                    }
                except asyncio.TimeoutError:
                    return {
                        "msg": "recv timeout",
                        "error": 5
                    }

            # Chain together connections -- fully async.
            p = req.api("/p2p/pipe/([^/]*)", named)
            if p:
                # Check con exists.
                if p["con_name"] not in self.cons:
                    await pipe.close()
                    return None

                con = self.cons[ p["con_name"] ]

                # Remove this server handler from con.
                # This pipe is no longer for HTTP!
                pipe.del_msg_cb(self.msg_cb)

                # Forward messages from pipe to con.
                # pipe -> con
                pipe.add_pipe(con)

                # Forward messages from con to pipe.
                # con  -> pipe
                con.add_pipe(pipe)

                # con <-----> pipe 
                return None

            # Binary send / recv methods.
            p = req.api("/p2p/binary/([^/]*)", named, optional)
            if p:
                # Check con exists.
                if p["con_name"] not in self.cons:
                    return {
                        "msg": f"Con name {p['con_name']} not found.",
                        "error": 4,
                    }

                # Send binary data from octet-stream POST.
                con = self.cons[ p["con_name"] ]
                if req.command == "POST":
                    # Content len header must exist.
                    if "Content-Length" not in req.hdrs:
                        return {
                            "msg": "content len header in binary POST",
                            "error": 6
                        }

                    # Content len must not exceed msg len.
                    payload_len = to_n(req.hdrs["Content-Length"])
                    if not in_range(payload_len, [1, len(msg)]):
                        return {
                            "msg": "invalid content len for bin POST",
                            "error": 7
                        }

                    # Last content-len bytes == payload.
                    data = msg[-payload_len:]
                    await con.send(data, con.stream.dest_tup)

                    # Return status.
                    return {
                        "name": p["con_name"],
                        "sent": payload_len,
                        "error": 0
                    }

                # Get buffer and send as binary stream.
                if req.command == "GET":
                    set_defaults(p, optional, defaults)
                    timeout = to_n(p["timeout"])
                    sub = [ p["msg_p"], p["addr_p"] ]

                    # Get binary from matching buffer.
                    out = await con.recv(sub, timeout=timeout, full=True)
                    if out is None:
                        return {
                            "msg": f"recv buffer {sub} empty.",
                            "error": 6
                        }

                    # Send it if any.
                    if out is not None:
                        await send_binary(out, req, client_tup, pipe)
                        return None

            # Close a connection.
            p = req.api("/p2p/close/([^/]*)", named)
            if p:
                # Check con exists.
                if p["con_name"] not in self.cons:
                    return {
                        "msg": f"Con name {p['con_name']} not found.",
                        "error": 4,
                    }

                # Close the con -- fires cleanup handler.
                con = self.cons[ p["con_name"] ]
                await con.close()

                # Indicate closed.
                return {
                    "closed": p["con_name"],
                    "error": 0
                }

            return {
                "msg": "No API method found.",
                "error": 1
            }

        resp = await async_wrap_errors(get_response())
        if resp is not None:
            await send_json(
                resp,
                req,
                client_tup,
                pipe
            )

    async def close(self):
        await self.node.close()
        await super().close()

# pragma: no cover
async def start_p2pd_server(ifs=None, route=None, port=0, do_loop=True, do_init=True, enable_upnp=True):
    print("Loading interfaces...")
    print("If you've just connected a new NIC ")
    print("there can be a slight delay until it's online.")
    if enable_upnp:
        print("Doing node port forwarding and pin hole rules. Please wait.")

    # Load netifaces.
    netifaces = None
    if do_init:
        netifaces = await init_p2pd()

    # Start node server.
    ifs = ifs or await load_interfaces(netifaces=netifaces)
    #port = get_port_by_ip
    node = await start_p2p_node(
        # Attempt deterministic port allocation based on NICs.
        # If in use a random port will be used.
        port=NODE_PORT + 60 + 1,
        ifs=ifs,
        enable_upnp=enable_upnp
    )

    # Specify listen port details.
    if route is not None and port is not None:
        route = await route.bind(port=port)
    else:
        route = await ifs[0].route().bind(ips="127.0.0.1")

    # Start P2PD server.
    p2p_server = P2PDServer(ifs, node)
    await p2p_server.listen_all(
        [route],
        [port],
        [TCP]
    )

    # Stop this thread exiting.
    bind_tup = p2p_server.servers[0][2].sock.getsockname()
    print(f"Started server on http://{bind_tup[0]}:{bind_tup[1]}")
    if do_loop:
        while 1:
            await asyncio.sleep(1)

    return p2p_server

if __name__ == "__main__":
    async_test(start_p2pd_server)


from p2pd import *

class EchoServer(Daemon):
    def __init__(self):
        super().__init__()
    #
    async def msg_cb(self, msg, client_tup, pipe):
        await pipe.send(msg, client_tup)

async def example():
    i = await Interface().start()
    #
    # Daemon instance.
    server_port = 10126
    echod = await EchoServer().listen_all(
        [i],
        [server_port],
        [TCP]
    )
    #
    await echod.close()

if __name__ == '__main__':
    async_test(example)


import platform
from p2pd.test_init import *
from p2pd.utils import *
from p2pd.net import VALID_AFS
from p2pd.win_netifaces import *
from p2pd.echo_server import *
from p2pd.toxiclient import *
from p2pd.toxiserver import *

streams = lambda: [ToxiToxic().downstream(), ToxiToxic().upstream()]

class TestToxid(unittest.IsolatedAsyncioTestCase):
    async def do_setup(self):
        # Use this for socket tests.
        net_conf = dict_child({
            "recv_timeout": 10
        }, NET_CONF)

        # Create main Toxi server.
        i = await Interface().start_local(skip_resolve=True)
        route = await i.route().bind(ips="127.0.0.1")
        toxid = ToxiMainServer([i])
        await toxid.listen_specific(
            [[route, TCP]]
        )

        # Create a Toxi client.
        listen_port = toxid.get_listen_port()
        toxid_addr = await Address("127.0.0.1", listen_port, route)
        client = await ToxiClient(
            toxid_addr,
            net_conf
        ).start()

        # Create an echo server -- used for some tests.
        route = await i.route().bind(ips="127.0.0.1")
        echod = EchoServer()
        await echod.listen_specific(
            [[route, TCP]]
        )

        # Address to connect to echod.
        listen_port = echod.get_listen_port()
        echo_dest = await Address("127.0.0.1", listen_port, route)
        return net_conf, i, toxid, toxid_addr, client, echod, echo_dest

    """
    If using HTTP as the application protocol to test that
    toxics work as expected you definitely want a new tunnel each
    time since reusing the same connection for multiple requests
    makes this overly complex. At least with the code bellow
    the first toxic seems to work as expected.
    """
    async def new_tunnel(self, tunnel_dest, client):
        # Setup a tunnel to a host for testing.
        # The toxi tunnels upstream is Google.
        tunnel = await client.new_tunnel(tunnel_dest)
        assert(isinstance(tunnel, ToxiTunnel))
        return tunnel

    async def do_cleanup(self, toxid, echod):
        await toxid.close()
        await echod.close()

    async def test_add_latency(self):
        # Setup.
        net_conf, \
        i, \
        toxid, \
        toxid_addr, \
        client, \
        echod, \
        echo_dest = await self.do_setup()

        # Test downstream and upstream.
        lag_amount = 2000
        tunnel_dest = await Address("www.google.com", 80, i.route())
        for toxi_stream in streams():
            # Add the initial toxic to a new tunnel.
            tunnel = await self.new_tunnel(tunnel_dest, client)
            toxic = toxi_stream.add_latency(lag_amount)
            await tunnel.new_toxic(toxic)

            # HTTP get request to Google via a toxi tunnel.
            start_time = timestamp()
            curl = await tunnel.get_curl()
            ret = await curl.vars().get("/", conf=net_conf)
            end_time = timestamp()

            # Check data was received.
            assert(ret is not None)

            # Check lag amount.
            # It should lag.
            duration = end_time - start_time
            duration *= 1000 # Convert to ms.
            assert(duration >= (lag_amount * 0.8))

            # Check for timeout.
            timeout = (net_conf['recv_timeout'] * 1000) * 0.8
            assert(duration < timeout)

            # Cleanup.
            await tunnel.remove_toxic(toxic)
            await tunnel.close()

        await self.do_cleanup(toxid, echod)

    async def test_limit_data(self):
        # Setup.
        net_conf, \
        i, \
        toxid, \
        toxid_addr, \
        client, \
        echod, \
        echo_dest = await self.do_setup()

        send_buf = b"1" * 90
        for toxi_stream in streams():
            # Open new tunnel.
            tunnel = await self.new_tunnel(echo_dest, client)
            tunneld = d_vals(toxid.tunnel_servs)[0]
            up_sock = tunneld.upstream_pipe.sock

            # Add limit data toxic.
            toxic = toxi_stream.add_limit_data(100)
            await tunnel.new_toxic(toxic)
            pipe, _ = await tunnel.get_pipe()

            # We're allowed to send 90 bytes okay.
            await pipe.send(send_buf)
            recv_buf = await pipe.recv()
            assert(recv_buf == send_buf)

            # But 100 or over closes the socket.
            await pipe.send(send_buf)
            await asyncio.sleep(1)
            assert(
                # Downstream test.
                pipe.sock.fileno() == -1
                
                or

                # Upstream test.
                up_sock.fileno() == -1
            ) 

            # Cleanup.
            await pipe.close()
            await tunnel.remove_toxic(toxic)
            await tunnel.close()

        await self.do_cleanup(toxid, echod)

    async def test_reset_peer(self):
        # Setup.
        net_conf, \
        i, \
        toxid, \
        toxid_addr, \
        client, \
        echod, \
        echo_dest = await self.do_setup()

        for toxi_stream in streams():
            # Open new tunnel.
            tunnel = await self.new_tunnel(echo_dest, client)
            tunneld = d_vals(toxid.tunnel_servs)[0]
            up_sock = tunneld.upstream_pipe.sock

            # Add limit data toxic.
            toxic = toxi_stream.add_reset_peer(1000)
            await tunnel.new_toxic(toxic)
            pipe, _ = await tunnel.get_pipe()

            # Should close after timeout.
            assert(up_sock.fileno() != -1)
            assert(pipe.sock.fileno() != -1)
            await asyncio.sleep(1.1)
            await pipe.send(b'test')
            await asyncio.sleep(1)
            assert(
                # Downstream test.
                pipe.sock.fileno() == -1
                
                or

                # Upstream test.
                up_sock.fileno() == -1
            ) 

            # Cleanup.
            await pipe.close()
            await tunnel.remove_toxic(toxic)
            await tunnel.close()

        await self.do_cleanup(toxid, echod)

    async def test_timeout(self):
        # Setup.
        net_conf, \
        i, \
        toxid, \
        toxid_addr, \
        client, \
        echod, \
        echo_dest = await self.do_setup()

        net_conf = dict_child({
            "recv_timeout": 1
        }, NET_CONF)

        for n, toxi_stream in enumerate(streams()):
            # Open new tunnel.
            tunnel = await self.new_tunnel(echo_dest, client)
            tunneld = d_vals(toxid.tunnel_servs)[0]
            up_sock = tunneld.upstream_pipe.sock

            # Add limit data toxic.
            toxic = toxi_stream.add_timeout(2000)
            await tunnel.new_toxic(toxic)
            pipe, _ = await tunnel.get_pipe(net_conf)

            await pipe.send(b'test')
            await asyncio.sleep(1)
            out = await pipe.recv()
            assert(out is None)

            await asyncio.sleep(1.5)
            assert(
                # Downstream test.
                pipe.sock.fileno() == -1
                
                or

                # Upstream test.
                up_sock.fileno() == -1
            ) 

            # Cleanup.
            await pipe.close()
            await tunnel.remove_toxic(toxic)
            await tunnel.close()

        await self.do_cleanup(toxid, echod)

    async def test_slicer(self):
        # Setup.
        net_conf, \
        i, \
        toxid, \
        toxid_addr, \
        client, \
        echod, \
        echo_dest = await self.do_setup()

        send_buf = b"123456890" * 100
        for toxi_stream in streams():
            # Open new tunnel.
            tunnel = await self.new_tunnel(echo_dest, client)

            # Add slicer toxic.
            toxic = toxi_stream.add_slicer()
            await tunnel.new_toxic(toxic)
            pipe, _ = await tunnel.get_pipe()
            await pipe.send(send_buf)
            await asyncio.sleep(1)

            # Receive sliced packets.
            recv_lens = []; recv_buf = b"" 
            while len(recv_buf) != len(send_buf):
                buf = await pipe.recv()
                recv_buf += buf
                recv_lens.append(len(buf))

            # Assert packet no and recv buffers.
            assert(recv_buf == send_buf)
            avg_recv = sum(recv_lens) / len(recv_lens)
            assert(avg_recv <= 150 and avg_recv >= 50)

            # Cleanup.
            await pipe.close()
            await tunnel.remove_toxic(toxic)
            await tunnel.close()

        await self.do_cleanup(toxid, echod)

    async def test_bandwidth_limit(self):
        # Setup.
        net_conf, \
        i, \
        toxid, \
        toxid_addr, \
        client, \
        echod, \
        echo_dest = await self.do_setup()

        bw_limit = 1; kb_send = 3; test_no = 3
        send_buf = b"1234567890" * 15000
        send_buf = send_buf[:1024 * kb_send]
        assert(len(send_buf) == 1024 * kb_send)
        for toxi_stream in streams():
            # Open new tunnel.
            tunnel = await self.new_tunnel(echo_dest, client)

            # Add bw data toxic.
            toxic = toxi_stream.add_bandwidth_limit(bw_limit)
            await tunnel.new_toxic(toxic)
            pipe, _ = await tunnel.get_pipe()
            readings = []
            for _ in range(0, test_no):
                start_time = time.time()
                await pipe.send(send_buf)

                recv_buf = b""
                while len(recv_buf) != len(send_buf):
                    recv_buf += await pipe.recv()

                end_time = time.time()
                duration = end_time - start_time
                readings.append(duration)

            # This is average to send 5 kbs.
            # Reducing the average by 5 is the average for 1 kb.
            # How closely does that match the bandwidth limit?
            avg_duration = sum(readings) / len(readings)
            avg_kb_rate = kb_send / avg_duration
            assert(avg_kb_rate <= bw_limit)
            assert(avg_kb_rate >= (bw_limit * 0.7))

            # Cleanup.
            await pipe.close()
            await tunnel.remove_toxic(toxic)
            await tunnel.close()

        await self.do_cleanup(toxid, echod)
            
    async def test_slow_close(self):
        # Setup.
        net_conf, \
        i, \
        toxid, \
        toxid_addr, \
        client, \
        echod, \
        echo_dest = await self.do_setup()

        test_time = 1500
        for n, toxi_stream in enumerate(streams()):
            # Open new tunnel.
            tunnel = await self.new_tunnel(echo_dest, client)
            tunneld = d_vals(toxid.tunnel_servs)[0]

            # Add limit data toxic.
            toxic = toxi_stream.add_slow_close(test_time)
            await tunnel.new_toxic(toxic)
            pipe, _ = await tunnel.get_pipe()

            # Should close after timeout.
            await pipe.send(b'test')
            await asyncio.sleep(1)

            start_time = time.time()
            if n == 0:
                client_pipe = tunneld.servers[0][2].tcp_clients[0]
                await client_pipe.close()
            else:        
                await tunneld.upstream_pipe.close()

            end_time = time.time()
            duration = end_time - start_time
            assert(duration * 1000 >= (test_time * 0.7))

            # Cleanup.
            await pipe.close()
            await tunnel.remove_toxic(toxic)
            await tunnel.close()

        await self.do_cleanup(toxid, echod)


if __name__ == '__main__':

    main()



"""
nc -4 -u p2pd.net 7

"""

from p2pd.test_init import *
try:
    from .static_route import *
except:
    from static_route import *
from p2pd.utils import what_exception, log_exception
from p2pd import IPRange, Bind, Route, Interface, pipe_open
from p2pd import Address
from p2pd.net import BLACK_HOLE_IPS, ip_norm, DUEL_STACK, IP6, IP4
from p2pd.net import NIC_BIND, EXT_BIND, VALID_AFS, TCP, UDP
from p2pd.base_stream import SUB_ALL

asyncio.set_event_loop_policy(SelectorEventPolicy())

class TestAFsWork(unittest.IsolatedAsyncioTestCase):
    async def test_afs(self):
        # List of public echo servers.
        addr = {
            UDP: {
                IP4: {
                    "host": "52.43.121.77",
                    "port": 10001
                },
                # TODO: Find IPv6 UDP public echo server.
            },
            TCP: {
                IP4: {
                    "host": "tcpbin.com",
                    "port": 4242
                },
                IP6: {
                    "host": "p2pd.net",
                    "port": 7
                }
            }
        }

        # Test all available AFs + protos.
        one_worked = False
        for af in VALID_AFS:
            try:
                # Get default Interface for AF type.
                i = Interface(af)
                rp = use_fixed_rp(i)
                await i.start(rp)
                one_worked = True
            except Exception:
                # Skip test if not supported.
                continue

            try:
                # Echo server address.
                route = await i.route(af).bind()

                # Test echo server with AF.
                msg = b"echo test"
                for proto in [TCP, UDP]:
                    # No server for this AF + proto.
                    if af not in addr[proto]:
                        continue

                    # Set destination of echo server.
                    echo_dest = await Address(
                        addr[proto][af]["host"],
                        addr[proto][af]["port"],
                        route
                    ).res()

                    # Open pipe to echo server.
                    pipe = await pipe_open(proto, route, echo_dest)
                    
                    # Interested in any message.
                    pipe.subscribe(SUB_ALL)

                    # Send data down the pipe.
                    for i in range(0, 4):
                        await pipe.send(msg + b"\r\n", echo_dest.tup)

                    # Receive data back.
                    data = await pipe.recv(SUB_ALL, 4)
                    assert(msg in data)

                    # Cleanup.
                    await pipe.close()
            except:
                continue

        self.assertTrue(one_worked)

if __name__ == '__main__':
    main()


from p2pd import *

TURN_OFFSET = 0

async def example():
    # Network interface details.
    i = await Interface().start()
    r = await i.route().bind()
    #
    # Address of a TURN server.
    dest = await Address(
        TURN_SERVERS[TURN_OFFSET]["host"],
        TURN_SERVERS[TURN_OFFSET]["port"],
        r
    ).res()
    #
    # Sync message callback -- do something here if you like.
    # Can be async too.
    msg_cb = lambda msg, client_tup, pipe: print(msg)
    #
    # Implement the TURN protocol for UDP send / recv.
    client = TURNClient(
        turn_addr=dest,
        turn_user=TURN_SERVERS[TURN_OFFSET]["user"],
        turn_pw=TURN_SERVERS[TURN_OFFSET]["pass"],
        turn_realm=TURN_SERVERS[TURN_OFFSET]["realm"],
        route=r,
        msg_cb=msg_cb
    )
    #
    # Wait for authentication and relay address allocation.
    await async_wrap_errors(
        client.start()
    )
    #
    # Give this to a client to send to ourselves.
    our_relay_tup = await client.relay_tup_future
    our_client_tup = await client.client_tup_future
    #
    """
    To receive messages back from a given client you will have
    to call: await client.accept_peer(their_client_tup, their_relay_tup) which implies you have your own way to
    exchange these details between clients (I use MQTT.)

    Example (after exchanging relay tups):
        await client_a.send(b'hello b', client_b_relay_tup)
        await client_b.send(b"hello a", client_a_relay_tup)

    To receive a message you would do:
        await client_a.recv(client_b_relay_tup)
    """
    # Cleanup.
    await client.close()

if __name__ == '__main__':
    async_test(example)


from p2pd import *

async def example():
    i = await Interface().start()
    addr = await Address("www.example.com", 80, i.route())
    curl = WebCurl(addr)
    params = {"action": "upload"}
    payload = b"Data to POST!"
    resp = await curl.vars(
        url_params=params,
        body=payload
    ).post("/meow/cat.php")
    print(resp.out)

if __name__ == '__main__':
    async_test(example)


import asyncio
import unittest
import platform
import socket
import hashlib
from unittest import main
from os import environ
import sys

from .errors import *
from .settings import *
from .utils import *
from .cmd_tools import *
from .net import *
from .address import *
from .interface import *

# Loads interface info on Windows.
# Make it available for all tests.
from .base_stream import *
from .stun_client import *


P2PD_NET_ADDR_BYTES = b'0,3-[0,149.56.128.148,149.56.128.148,10001,1,1,0]-[0,2607:5300:0201:3100:0000:0000:0000:8d2f,fe80:0000:0000:0000:f816:3eff:feae:b2d9,10001,1,1,0]-p2pd_test_node'

# Only load the test interface on the right machine.
# Otherwise the name (probably) won't exist.
P2PD_IFS = []

vmaj, vmin, _ = platform.python_version_tuple()
vmaj = int(vmaj); vmin = int(vmin)
if vmaj < 3:
    raise Exception("Python 2 not supported.")
if vmin <= 4:
    raise Exception("Project needs >= 3.5")

if not hasattr(unittest, "IsolatedAsyncioTestCase"):
    log("patching isolated asyncio test case")
    import aiounittest
    unittest.IsolatedAsyncioTestCase = aiounittest.AsyncTestCase


class PatchedAsyncTest(unittest.IsolatedAsyncioTestCase):
    def _setupAsyncioLoop(self):
        assert self._asyncioTestLoop is None
        loop = asyncio.new_event_loop()

        asyncio.set_event_loop(loop)
        loop.set_debug(False)
        self._asyncioTestLoop = loop
        fut = loop.create_future()
        self._asyncioCallsTask = loop.create_task(self._asyncioLoopRunner(fut))
        loop.run_until_complete(fut)

# Basic echo client test.
async def check_pipe(pipe, dest_tup=None):
    # Sanity check.
    data = b"Meow"
    if pipe is None:
        log("> Check pipe: pipe is none")
        return False

    # Indicate any message is acceptable to queue.
    pipe.subscribe(sub=SUB_ALL)

    # Request peer to echo back the data sent.
    if dest_tup is None:
        dest_tup = pipe.sock.getpeername()

    # Give any pipes time to add callback handlers.
    # This prevents race conditions in processing messages.
    await asyncio.sleep(1)
    await pipe.echo(data, dest_tup)

    # Wait for message with timeout.
    buf = await pipe.recv(timeout=3)
    if buf is None:
        log("Check pipe: recv buf is none")
        return False

    # May not have echo process hooked up.
    if data in buf:
        return True
    else:
        log(f"Check pipe: invalid recv buf {buf}")
        return False

# If a random node ID is generated and subbed to over and over.
# The server may limit new subs from the same IP.
# But if all test nodes use the same IDs the there will be collisons.
# Hence generate a unique IQ deterministically.
def node_name(x, i):
    assert(i.resolved)
    name_base = to_b(f"{i.mac} {socket.gethostname()}")
    node_name = hashlib.sha256(x + name_base).hexdigest()
    return to_b(node_name)[:10]

class FakeSTUNClient():
    def __init__(self, interface=None, af=IP4):
        self.interface = interface
        self.rip = "1.3.3.7"
        self.sock = None
        self.mappings = [] # [local, mapped] ...
        self.p = 0
        self.wan_ip = None
        self.af = af

    def rand_server(self):
        return None

    def set_mappings(self, mappings):
        self.mappings = mappings
        self.p = 0

    def set_wan_ip(self, wan_ip):
        self.wan_ip = wan_ip

    async def get_wan_ip(self, af=None, interface=None, fast_fail=0, servers=None, local_addr=None, conf=STUN_CONF):
        return ip_norm(self.wan_ip)

    async def get_mapping(self, proto, af=None, source_port=0, group="map", alt_port=0, do_close=0, fast_fail=0, servers=None, conf=STUN_CONF):
        run_time = time.time()
        local, mapped = self.mappings[self.p]
        out = [self.interface, self.sock, local, mapped, self.rip, run_time]
        self.p = (self.p + 1) % len(self.mappings)

        return out
    
async def duel_if_setup(netifaces, load_nat=False):
    # Load interface list.
    ifs = await load_interfaces(netifaces, load_nat=load_nat)
    ifs, af = get_ifs_by_af_intersect(ifs)
    if len(ifs) <= 1:
        return [], None
    
    # Only need at most 2.
    ifs = ifs[:2]
    if(ifs[0].route(af).ext() != ifs[1].route(af).ext()):
        return ifs, af
    else:
        return [], None

class FakeNetifaces():
    def __init__(self):
        self.addr_info = None

    def set_addr_info(self, addr_info):
        self.addr_info = addr_info

    def ifaddresses(self, if_name):
        return self.addr_info


from p2pd.test_init import *

import warnings
import socket
import multiprocessing
from decimal import Decimal as Dec
from p2pd.utils import *
from p2pd.p2p_addr import is_p2p_addr_us
from p2pd.p2p_node import *
from p2pd.p2p_utils import *
from p2pd.p2p_pipe import *
from p2pd.interface import *

asyncio.set_event_loop_policy(SelectorEventPolicy())
async def test_setup(netifaces=None, ifs=None):
    # Suppress socket unclosed warnings.
    if not hasattr(test_setup, "netifaces"):
        if netifaces is None:
            test_setup.netifaces = await init_p2pd()
        else:
            test_setup.netifaces = netifaces
        warnings.filterwarnings('ignore', message="unclosed", category=ResourceWarning)

    # Static setup because it's fast.
    if not hasattr(test_setup, "ifs"):
        if socket.gethostname() == "p2pd.net":
            test_setup.ifs = P2PD_IFS
        else:
            # Load a list of interfaces to use for the tests.
            test_setup.ifs = await load_interfaces(netifaces=test_setup.netifaces)

    # Load clock skew for test machine.
    if not hasattr(test_setup, "clock_skew"):
        test_setup.clock_skew = (await SysClock(test_setup.ifs[0]).start()).clock_skew

    # Load process pool executors.
    pp_executors = await get_pp_executors(workers=2)

    # Main node used for testing p2p functionality.
    node_a_ifs = [ifs[0]] if ifs is not None else test_setup.ifs
    node_a = await start_p2p_node(
        node_id=node_name(b"node_a_" + rand_plain(8), node_a_ifs[0]),

        # Get brand new unassigned listen port.
        # Avoid TIME_WAIT buggy sockets from port reuse.
        port=0,
        ifs=node_a_ifs,
        clock_skew=test_setup.clock_skew,
        pp_executors=pp_executors,
        enable_upnp=False
    )

    # Test local punching algorithm.
    node_b_ifs = [ifs[1  % len(ifs)]] if ifs is not None else test_setup.ifs
    node_b = await start_p2p_node(
        node_id=node_name(b"node_b_" + rand_plain(8), node_b_ifs[0]),

        # Get brand new unassigned listen port.
        # Avoid TIME_WAIT buggy sockets from port reuse.
        port=0,
        ifs=node_b_ifs,
        clock_skew=test_setup.clock_skew,
        pp_executors=pp_executors,
        enable_upnp=False
    )

    print(f"Node a = {node_a.p2p_addr}")
    print(f"Node b = {node_b.p2p_addr}")
    return node_a, node_b

async def do_node_cleanup(node_a, node_b):
    await node_a.close()
    await node_b.close()

class TestP2PPipe(unittest.IsolatedAsyncioTestCase):
    # Self connect won't work due to Coturn changes.
    async def test_self_turn_connect(self):
        return
        log(">>> test_self_turn_connect")
        node_a, node_b = await test_setup()
        pipe, _ = await node_a.connect(
            node_b.address(),
            strategies=[P2P_RELAY]
        )

        self.assertTrue(pipe is not None)
        dest_tup = await pipe.relay_tup_future
        pipe_okay = await check_pipe(pipe, dest_tup=dest_tup)
        self.assertTrue(pipe_okay)
        await pipe.close()
        await do_node_cleanup(node_a, node_b)

    
    async def test_multiple_strats(self):
        # Test reverse connect.
        log(">>> test_multiple_strats")
        node_a, node_b = await test_setup()

        # Replace node b's port with an invalid one.
        node_b_addr = node_b.address()
        node_b_server = node_b.servers[0][2]
        node_b_port = str(node_b_server.sock.getsockname()[1])
        node_b_addr.replace(to_b(node_b_port), b"68000")

        # Connect using different strategies.
        pipe, _ = await node_a.connect(
            node_b_addr,
            strategies=[
                # Direct will fail.
                P2P_DIRECT,
                
                # Reverse should succeed.
                P2P_REVERSE
            ]
        )
        self.assertTrue(pipe is not None)

        # Test pipe is valid.
        pipe_okay = await check_pipe(pipe)
        self.assertTrue(pipe_okay)
        await pipe.close()
        await do_node_cleanup(node_a, node_b)

    async def test_self_reverse_connect(self):
        # Test reverse connect.
        log(">>> test_self_reverse_connect")
        node_a, node_b = await test_setup()
        pipe, _ = await node_a.connect(
            node_b.address(),
            strategies=[P2P_REVERSE]
        )
        self.assertTrue(pipe is not None)

        # Test pipe is valid.
        pipe_okay = await check_pipe(pipe)
        self.assertTrue(pipe_okay)
        await pipe.close()
        await do_node_cleanup(node_a, node_b)

    async def test_self_direct_connect(self):
        # Test reverse connect.
        log(">>> test_self_direct_connect")
        node_a, node_b = await test_setup()
        pipe, _ = await node_a.connect(
            node_b.address(),
            strategies=[P2P_DIRECT]
        )
        self.assertTrue(pipe is not None)

        # Test pipe is valid.
        pipe_okay = await check_pipe(pipe)
        self.assertTrue(pipe_okay)
        await pipe.close()
        await do_node_cleanup(node_a, node_b)

    """
    async def test_remote_direct_connect(self):
        if not P2PD_TEST_INFRASTRUCTURE:
            return

        # Test direct connect
        log(">>> test_remote_direct_connect")
        node_a, node_b = await test_setup()
        pipe, _ = await node_a.connect(
            P2PD_NET_ADDR_BYTES,
            strategies=[P2P_DIRECT]
        )
        self.assertTrue(pipe is not None)

        # Test pipe is valid.
        pipe_okay = await check_pipe(pipe)
        self.assertTrue(pipe_okay)
        await pipe.close()
        await do_node_cleanup(node_a, node_b)
    """

    async def test_self_punch(self):
        # Test p2p punch to self.
        # This also tests local punching as the same code is used.
        log(">>> test_self_punch")
        node_a, node_b = await test_setup()
        pipe, _ = await node_a.connect(
            node_b.address(),
            strategies=[P2P_PUNCH]
        )

        # Test pipe is valid.
        pipe_okay = await check_pipe(pipe)
        if not pipe_okay:
            print("Test self punch with networking failed.")
            print("Test is optional but mentioning it.")

        if pipe is not None:
            await pipe.close()
        await do_node_cleanup(node_a, node_b)

    """
    async def test_remote_reverse_connect(self):
        if not P2PD_TEST_INFRASTRUCTURE:
            return

        # Test reverse connect.
        log(">>> test_remote_reverse_connect")
        node_a, node_b = await test_setup()
        found_open = False
        for interface in node_a.if_list:
            if interface.nat["type"] == OPEN_INTERNET:
                pipe, _ = await node_a.connect(
                    P2PD_NET_ADDR_BYTES,
                    strategies=[P2P_REVERSE]
                )
                self.assertTrue(pipe is not None)

                # Test pipe is valid.
                pipe_okay = await check_pipe(pipe)
                self.assertTrue(pipe_okay)
                await pipe.close()
                found_open = True
                break

        if not found_open:
            print("> Skipping reverse connect test")
            print("> NAT type is not open.\r\n")

        await do_node_cleanup(node_a, node_b)

    async def test_remote_punch_duel_ifs(self):
        # Load interface list.
        netifaces = await init_p2pd()
        ifs, af = await duel_if_setup(netifaces, load_nat=True)
        if af is None:
            return
        
        # Start nodes on different interfaces.
        node_a, node_b = await test_setup(netifaces, ifs)
        pipe, _ = await node_a.connect(
            node_b.address(),
            strategies=[P2P_PUNCH]
        )

        # Check if a connection was spawned.
        self.assertTrue(pipe is not None)
        pipe_okay = await check_pipe(pipe)
        self.assertTrue(pipe_okay is not None)

        # Cleanup        
        if pipe is not None:
            await pipe.close()
        await do_node_cleanup(node_a, node_b)

    async def test_remote_punch(self, netifaces=None, ifs=None, is_optional=False):
        if not P2PD_TEST_INFRASTRUCTURE and ifs is None:
            return

        # If P2PD_NET_ADDR_BYTES is not us then test punching to it.
        log(">>> test_remote_punch")
        node_a, node_b = await test_setup(netifaces, ifs)
        if not is_p2p_addr_us(P2PD_NET_ADDR_BYTES, node_a.if_list):
            # Connect to p2pd.net test server.
            pipe, _ = await node_a.connect(
                P2PD_NET_ADDR_BYTES,
                strategies=[P2P_PUNCH]
            )
            self.assertTrue(pipe is not None)

            # Test pipe is valid.
            pipe_okay = await check_pipe(pipe)
            if not pipe_okay and is_optional:
                log("remote punch failed but was optional.")
            else:
                self.assertTrue(pipe_okay)
            await pipe.close()
        else:
            print("> p2pd net addr is us. Skipping remote punch test.")

        await do_node_cleanup(node_a, node_b)

    async def test_remote_turn_connect(self):
        if not P2PD_TEST_INFRASTRUCTURE:
            return

        log(">>> test_remote_turn_connect")
        node_a, node_b = await test_setup()
        pipe, _ = await node_a.connect(
            P2PD_NET_ADDR_BYTES,
            strategies=[P2P_RELAY]
        )

        self.assertTrue(pipe is not None)

        # Get the relay address of the test peer.
        p2pd_ip = list(pipe.peers.keys())[0]
        dest_tup = pipe.get_relay_tup(p2pd_ip)

        # Use it to do the echo test on TURN client.
        # Use the peers relay address as the destination.
        pipe_okay = await check_pipe(pipe, dest_tup=dest_tup)
        self.assertTrue(pipe_okay)
        await pipe.close()
        await do_node_cleanup(node_a, node_b)
    """

if __name__ == '__main__':
    main()




from p2pd import *

class CustomServer(RESTD):
    @RESTD.GET()
    async def show_index(self, v, pipe):
        print(v)
        #
        # text/html response.
        return "hello"
    #
    @RESTD.POST(["proxies"], ["toxics"])
    async def add_new_toxic(self, v, pipe):
        """
        Matches /proxies/.*/toxics/.*
        and names the value at 
        v['proxies'] and v['toxics']
        respectively.
        """
        #
        # Demonstrates a JSON response.
        return {
            "status": "success"
        }
    #
    @RESTD.DELETE(["proxies"], ["toxics"])
    async def del_new_toxic(self, v, pipe):
        # Application/octet-stream response.
        return b""

async def example():
    i = await Interface().start_local()
    server = CustomServer()
    await server.listen_all(
        [
            await i.route(IP4),
            await i.route(IP4).bind(ips="127.0.0.1")
        ],
        [60322],
        [TCP, UDP],
    )
    #
    """
    Feel free to add a while: sleep
    then test it with CURL or something...
    """
    #
    await server.close()

if __name__ == '__main__':
    async_test(example)


import sys
import re
import platform
import multiprocessing
import socket
import pprint
from .errors import *
from .settings import *
from .route import *
from .nat import *
from .route_table import *
if sys.platform == "win32":
    from .win_netifaces import *
else:
    import netifaces as netifaces


async def init_p2pd():
    global ENABLE_UDP
    global ENABLE_STUN

    # Setup event loop.
    loop = asyncio.get_event_loop()
    loop.set_debug(False)
    #loop.set_exception_handler(SelectorEventPolicy.exception_handler)
    
    def fatal_error(self, exc, message='Fatal error on transport'):
        # Should be called from exception handler only.
        self._loop.call_exception_handler({
            'message': message,
            'exception': exc,
            'transport': self,
            'protocol': self._protocol,
        })
        self._force_close(exc)

    #asyncio.selector_events._SelectorTransport._fatal_error = fatal_error

    # Attempt to get monkey patched netifaces.
    netifaces = Interface.get_netifaces()
    if netifaces is None:
        if sys.platform == "win32":
            """
            loop = get_running_loop()

            # This happens if the asyncio REPL is used.
            # Nested event loops are a work around.
            if loop is not None:
                import nest_asyncio
                nest_asyncio.apply()
            """
            netifaces = await Netifaces().start()
        else:
            netifaces = sys.modules["netifaces"]

        Interface.get_netifaces = lambda: netifaces

    # Are UDP sockets blocked?
    # Firewalls like iptables on freehosts can do this.
    sock = None
    try:
        # Figure out what address family default interface supports.
        _, af = get_default_iface(netifaces)
        if af == AF_ANY: # Duel stack. Just use v4.
            af = IP4

        # Set destination based on address family.
        if af == IP4:
            dest = ('8.8.8.8', 60000)
        else:
            dest = ('2001:4860:4860::8888', 60000)

        # Build new UDP socket.
        sock = socket.socket(family=af, type=socket.SOCK_DGRAM)

        # Attempt to send small msg to dest.
        sock.sendto(b'testing UDP. disregard this sorry.', 0, dest)
    except Exception:
        """
        Maybe in the future I write code as a fail-safe but for
        now I don't have time. It's better to show a clear reason
        why the library won't work then to silently fail.
        """
        raise Exception("Error this library needs UDP support to work.")
    

        ENABLE_UDP = False
        ENABLE_STUN = False
        log("UDP sockets blocked! Disabling STUN.")
        log_exception()
    finally:
        if sock is not None:
            sock.close()

    return netifaces

def get_default_iface(netifaces, preference=AF_ANY, exp=1, duel_stack_test=True):
    gws = netiface_gateways(netifaces, get_interface_type, preference=preference)
    gateway = None
    iface = None
                    
    # Convert any to netifaces.af.
    if preference == AF_ANY:
        # First valid address family for default gateway.
        preference = list(gws["default"])[0]
    else:
        preference = af_to_netiface(preference)
    
    # Get address of gateway for 'default' interface.
    preference = int(preference)
    if preference in gws["default"]:
        iface = gws["default"][preference][1]
            
    # Check found a default interface for ipv4 or ipv6.
    if iface == None:
        raise Exception("Couldn't find default WAN interface. Specify IPs manually in config to bypass this error.")

    # See if this iface is duel-stack.
    af = netiface_to_af(preference, netifaces)
    if duel_stack_test:
        af_index = VALID_AFS.index(preference)
        af_len = len(VALID_AFS)
        other_af = VALID_AFS[(af_index - 1) % af_len]
        try:
            other_iface, _ = get_default_iface(netifaces, other_af, duel_stack_test=False)
        except Exception:
            other_iface = None
        if other_iface == iface:
            af = DUEL_STACK
        
    return [iface, af]

def get_interface_type(name):
    name = name.lower()
    if re.match("en[0-9]+", name) != None:
        return INTERFACE_ETHERNET

    eth_names = ["eth", "eno", "ens", "enp", "enx", "ethernet"]
    for eth_name in eth_names:
        if eth_name in name:
            return INTERFACE_ETHERNET

    wlan_names = ["wlp", "wireless", "wlan", "wifi"]
    for wlan_name in wlan_names:
        if wlan_name in name:
            return INTERFACE_WIRELESS

    if "wl" == name[0:2]:
        return INTERFACE_WIRELESS

    return INTERFACE_UNKNOWN

def get_interface_stack(rp):
    stacks = []
    for af in [IP4, IP6]:
        if af in rp:
            if len(rp[af].routes):
                stacks.append(af)

    if len(stacks) == 2:
        return DUEL_STACK

    if len(stacks):
        return stacks[0]

    return UNKNOWN_STACK

# Used for specifying the interface for sending out packets on
# in TCP streams and UDP streams.
class Interface():
    def __init__(self, name=None, stack=DUEL_STACK, nat=None, netifaces=None):
        super().__init__()
        self.resolved = False
        self.nat = nat or nat_info()
        self.name = name
        self.rp = {IP4: RoutePool(), IP6: RoutePool()}
        self.v4_lan_ips = []
        self.guid = None
        self.netifaces = netifaces or Interface.get_netifaces()

        # Check NAT is valid if set.
        if nat is not None:
            assert(isinstance(nat, dict))
            assert(nat.keys() == nat_info().keys())

        # Can provide a stack type to skip processing unsupported AFs.
        # Otherwise all AFs are checked when start() is called.
        self.stack = stack
        assert(self.stack in VALID_STACKS)

    # Load mac, nic_no, and process name.
    def load_if_info(self):
        # Assume its an AF.
        if isinstance(self.name, int):
            if self.name not in [IP4, IP6, AF_ANY]:
                raise InterfaceInvalidAF

            self.name, _ = get_default_iface(self.netifaces, preference=self.name)

        # No name specified.
        # Get name of default interface.
        if self.name is None or self.name == "":
            # Windows -- default interface name is a GUID.
            # This is ugly AF.
            iface_name, iface_af = get_default_iface(self.netifaces)
            if iface_name is None:
                raise InterfaceNotFound
            else:
                self.name = iface_name

            # Allow blank interface names to be used for testing.
            log("> default interface loaded")

            # May not be accurate.
            # Start() is the best way to set this.
            if self.stack == DUEL_STACK:
                self.stack = iface_af

        # Windows NIC descriptions are used for the name
        # if the interfaces are detected as all hex.
        # It's more user friendly.
        self.name = to_s(self.name)

        # Check ID exists.
        if self.netifaces is not None:
            if self.name not in self.netifaces.interfaces():
                log(f"interface name {self.name} not in {self.netifaces.interfaces()}")
                raise InterfaceNotFound
            self.type = get_interface_type(self.name)
            self.nic_no = 0
            if hasattr(self.netifaces, 'nic_no'):
                self.nic_no = self.netifaces.nic_no(self.name)
                self.id = self.nic_no
            else:
                self.id = self.name

            # Set MAC address of Interface.
            self.mac = get_mac_address(self.name, self.netifaces)
            if self.mac is None:
                raise Exception("mac none.")

    def to_dict(self):
        from .var_names import TXT
        return {
            "name": self.name,
            "nic_no": self.nic_no,
            "id": self.id,
            "mac": self.mac,
            "is_default": {
                int(IP4): self.is_default(IP4),
                int(IP6): self.is_default(IP6)
            },
            "nat": {
                "type": self.nat["type"],
                "nat_info": TXT["nat"][ self.nat["type"] ],
                "delta": self.nat["delta"],
                "delta_info": TXT["delta"][ self.nat["delta"]["type"] ]
            },
            "rp": {
                int(IP4): self.rp[IP4].to_dict(),
                int(IP6): self.rp[IP6].to_dict()
            }
        }

    @staticmethod
    def get_netifaces():
        return None

    @staticmethod
    def list():
        return Interface.get_netifaces().interfaces()

    @staticmethod
    def from_dict(d):
        i = Interface(d["name"])
        i.nic_no = d["nic_no"]
        i.id = d["id"]
        i.mac = d["mac"]
        def is_default_wrapper(af, gws=None):
            return d["is_default"][af]
        i.is_default = is_default_wrapper

        # Set the interface route pool.
        i.rp = {
            IP4: RoutePool.from_dict(d["rp"][int(IP4)]),
            IP6: RoutePool.from_dict(d["rp"][int(IP6)])
        }

        # Set interface part of routes.
        for af in VALID_AFS:
            for route in i.rp[af].routes:
                route.interface = i

        # Set NAT details.
        i.nat = nat_info(d["nat"]["type"], d["nat"]["delta"])

        # Set stack type of the interface based on the route pool.
        i.stack = get_interface_stack(i.rp)

        # Indicate the interface is fully resolved.
        i.resolved = True

        # ... and return it.
        return i

    # Make this interface printable because it's useful.
    def __str__(self):
        return pprint.pformat(self.to_dict())

    # Show a representation of this object.
    def __repr__(self):
        return f"Interface.from_dict({str(self)})"

    # Pickle.
    def __getstate__(self):
        return self.to_dict()

    # Unpickle.
    def __setstate__(self, state):
        o = self.from_dict(state)
        self.__dict__ = o.__dict__

    async def start_local(self, rp=None, skip_resolve=False):
        # Load internal interface details.
        if Interface.get_netifaces() is None:
            self.netifaces = await init_p2pd()

        # Process interface name in right format.
        self.load_if_info()

        # Get routes for AF.
        if rp == None:
            af_list = self.supported(skip_resolve=1)
            tasks = []
            for af in af_list:
                async def helper(af):
                    route = await self.route_test(af)
                    if route is None:
                        # Empty route pool.
                        self.rp[af] = RoutePool()
                    else:
                        self.rp[af] = RoutePool([route])

                tasks.append(helper(af))

            # Get all the routes concurrently.
            await asyncio.gather(*tasks)
        else:
            self.rp = rp

        # Update stack type based on routable.
        self.stack = get_interface_stack(self.rp)
        assert(self.stack in VALID_STACKS)
        self.resolved = True
        return self

    async def start(self, rp=None, skip_resolve=False):
        # Load internal interface details.
        if Interface.get_netifaces() is None:
            self.netifaces = await init_p2pd()

        # Process interface name in right format.
        self.load_if_info()

        # Get routes for AF.
        if rp == None:
            af_list = self.supported(skip_resolve=1)
            tasks = []
            for af in af_list:
                async def helper(af):
                    try:
                        self.rp[af] = await Routes(
                            [self],
                            af,
                            self.netifaces,
                            skip_resolve
                        )
                    except NoGatewayForAF:
                        # Empty route pool.
                        self.rp[af] = RoutePool()
                        log(f"No route for gw {af}")
                
                tasks.append(
                    asyncio.wait_for(
                        async_wrap_errors(
                            helper(af)
                        ),
                        15
                    )
                )

            # Get all the routes concurrently.
            try:
                await asyncio.gather(*tasks)
            except asyncio.TimeoutError:
                raise Exception("Could not start iface in 15s.")
        else:
            self.rp = rp

        # Update stack type based on routable.
        self.stack = get_interface_stack(self.rp)
        assert(self.stack in VALID_STACKS)
        self.resolved = True
        return self

    def __await__(self):
        return self.start().__await__()

    def set_nat(self, nat):
        assert(isinstance(nat, dict))
        assert(nat.keys() == nat_info().keys())
        self.nat = nat

    async def load_nat(self):
        # Try to avoid circular imports.
        from .base_stream import pipe_open
        from .stun_client import STUNClient, STUN_CONF
        from .nat_test import fast_nat_test
        
        # IPv6 only has no NAT!
        if IP4 not in self.supported():
            nat = nat_info(OPEN_INTERNET, EQUAL_DELTA)
        else:
            # STUN is used to get the delta type.
            af = IP4
            stun_client = STUNClient(
                self,
                af
            )

            # Get the NAT type.
            route = await self.route(af).bind()
            pipe = await pipe_open(UDP, route=route, conf=STUN_CONF)

            # Run delta test.
            nat_type, delta = await asyncio.wait_for(
                asyncio.gather(*[
                    async_wrap_errors(
                        fast_nat_test(pipe, STUND_SERVERS[af])
                    ),
                    async_wrap_errors(
                        delta_test(stun_client)
                    )
                ]),
                timeout=4
            )

            nat = nat_info(nat_type, delta)
            await pipe.close()

        # Load NAT type and delta info.
        # On a server should be open.
        self.set_nat(nat)
        return nat

    def get_scope_id(self):
        assert(self.resolved)

        # Interface specified by no on windows.
        if platform.system() == "Windows":
            return self.nic_no
        else:
            # Other platforms just use the name
            return self.name

    def nic(self, af):
        # Sanity check.
        if self.resolved:
            assert(af in self.what_afs())
        if self.rp != {} and len(self.rp[af].routes):
            return self.route(af).nic()

    def route(self, af=None, bind_port=0):
        # Sanity check.
        if self.resolved:
            af = af or self.supported()[0]
            assert(af in self.what_afs())

        # Main route is last.
        if af in self.rp:
            if len(self.rp[af].routes):
                return copy.deepcopy(self.rp[af].routes[-1])

        raise Exception(f"No route for {af} found.")

    async def route_test(self, af):
        # Return route with no external address info set.
        nic_ips = await get_nic_private_ips(self, af, self.netifaces)
        if not len(nic_ips):
            return None

        return Route(
            af=af,
            nic_ips=nic_ips,
            ext_ips=[IPRange(BLACK_HOLE_IPS[af])],
            interface=self
        )

    """
    Using a default list of gateways like this has a small
    performance advantage but the cost is if the interfaces list
    changes at run time like a wifi network disconnecting then
    the is_default function may give the incorrect result. There
    should be a way to detect loss of internet connection though.
    """
    def is_default(self, af, gws=None):
        af = af_to_netiface(af)
        gws = gws or netiface_gateways(self.netifaces, get_interface_type, preference=af)
        def_gws = gws["default"]
        if af not in def_gws:
            return False
        else:
            info = def_gws[af]
            if info[1] == self.name:
                return True
            else:
                return False

    def supported(self, skip_resolve=0):
        if not skip_resolve:
            assert(self.resolved)

        if self.stack == UNKNOWN_STACK:
            raise Exception("Unknown stack")

        if self.stack == DUEL_STACK:
            return [IP4, IP6]
        else:
            return [self.stack]

    def what_afs(self):
        assert(self.resolved)
        return self.supported()

"""
Given a list of interfaces returned from netifaces
or the win_netifaces module this code will filter the list
so that only interfaces that are used for the Internet remain.
Already done in win_netifaces. Uses route tables for Linux and Mac.
Other OS is based on the interface name (not that accurate.)
"""
async def filter_trash_interfaces(netifaces=None):
    netifaces = netifaces or Interface.get_netifaces()
    ifs = netifaces.interfaces()
    os_family = platform.system()

    # Interface list already well filtered by win_netifaces.py.
    if os_family == "Windows":
        return ifs

    # Use route table for these OS family to determine if interface
    # is useful or not.
    if os_family in ["Linux", "Darwin"]:
        tasks = []
        for if_name in ifs:
            async def worker(if_name):
                r = await is_internet_if(if_name)
                if r:
                    return if_name
                else:
                    return None

            tasks.append(worker(if_name))

        results = await asyncio.gather(*tasks)
        results = strip_none(results)
        
        """
        The 'is_interface_if' function depends on using the 'route' binary.
        If it does not exist then the code will fail and return no results.
        In this case default to name-based filtering of netifaces.
        """
        if len(results):
            return results

    # Otherwise use the interface type function.
    # Looks at common patterns for interface names (not accurate.)
    clean_ifs = []
    for if_name in ifs:
        if_type = get_interface_type(if_name)
        if if_type != INTERFACE_UNKNOWN:
            clean_ifs.append(if_name)

    return clean_ifs

def log_interface_rp(interface):
    for af in VALID_AFS:
        if not len(interface.rp[af].routes):
            continue

        route_s = str(interface.rp[af].routes)
        log(f"> AF {af} = {route_s}")
        log(f"> nic() = {interface.route(af).nic()}")
        log(f"> ext() = {interface.route(af).ext()}")

def get_ifs_by_af_intersect(if_list):
    largest = []
    af_used = None
    for af in VALID_AFS:
        hay = []
        for iface in if_list:
            if af in iface.supported():
                hay.append(iface)

        if len(hay) > len(largest):
            largest = hay
            af_used = af

    return [largest, af_used]

async def load_interfaces(netifaces=None, load_nat=True):
    netifaces = netifaces or Interface.get_netifaces()
    if netifaces is None:
        netifaces = await init_p2pd()

    # Get list of good interfaces with ::/0 or 0.0.0.0 routes.
    ifs = await filter_trash_interfaces(netifaces)
    ifs = to_unique(ifs)
    if ifs == []:
        # Something must have gone wrong so just use regular netifaces.
        ifs = netifaces.interfaces()

    # Start all interfaces.
    if_list = []
    tasks = []
    for if_name in ifs:
        if_info = str(netifaces.ifaddresses(if_name))
        log(f"Attempt to start if name {if_name}")
        log(f"Net iface results for that if = {if_info}")
        async def worker(if_name):
            try:
                interface = await Interface(if_name, netifaces=netifaces).start()
                try:
                    if load_nat:
                        await interface.load_nat()
                except Exception:
                    log("Failed to load nat for interface.")
                    # Just use the default NAT info.

                if_list.append(interface)
            except Exception:
                log_exception()
                return

        tasks.append(
            # Assume timeout = non-routable.
            worker(if_name)
        )

    await asyncio.gather(*tasks)

    # Filter any interfaces that have no routes.
    # This will filter out loopback and other crap interfaces.
    good_ifs = []
    for interface in if_list:
        for af in VALID_AFS:
            if len(interface.rp[af].routes):
                good_ifs.append(interface)
                break

    # Log interfaces and routes.
    log("> Loaded interfaces.")
    for if_no, interface in enumerate(good_ifs):
        log(f"> Routes for interface {if_no}:")
        log_interface_rp(interface)

    return good_ifs

# Given a list of Interface objs.
# Convert to dict and return a list.
def if_list_to_dict(if_list):
    dict_list = []
    for interface in if_list:
        d = interface.to_dict()
        dict_list.append(d)

    return dict_list

# Given a list of Interface dicts.
# Convert them back to Interfaces and return a list.
def dict_to_if_list(dict_list):
    if_list = []
    for d in dict_list:
        interface = Interface.from_dict(d)
        if_list.append(interface)

    return if_list
            
# Stores a list of interfaces for the machine.
class Interfaces():
    def __init__(self, interfaces=[]):
        self.started = False
        self.by_name = {}
        self.by_af = {
            AF_INET: [],
            AF_INET6: []
        }

        # Interfaces are added once we know their stack type.
        # which is know after they are 'started' to tell
        # which address family is routable to the Internet.
        for interface in interfaces:
            self.add(interface)

        if len(interfaces):
            self.started = True

    def add(self, interface):
        duel_stack = 1 if interface.stack == DUEL_STACK else 0
        if interface.stack == AF_INET or duel_stack:
            self.by_af[AF_INET].append(interface)

        if interface.stack == AF_INET6 or duel_stack:
            self.by_af[AF_INET6].append(interface)

        self.by_name[interface.name] = interface
        return interface

    def get(self, af):
        return self.by_af[af][0]

    # Load default interfaces.
    async def start(self, do_start=1): # pragma: no cover
        tasks = []

        # Load default interfaces.
        netifaces = Interface.get_netifaces()
        if self.names == {}:
            gws = netifaces.gateways()["default"]
            interface_one = None
            name_index = {}
            for af in [netifaces.AF_INET, netifaces.AF_INET6]:
                if af in gws:
                    name = gws["default"][af][1]
                    if name not in name_index:
                        interface = Interface(name)
                        if do_start:
                            tasks.append(
                                interface.start()
                            )

                        self.names[name] = interface
                        name_index[name] = 1

        # Get WAN IPs for interfaces concurrently.
        if len(tasks):
            await asyncio.gather(*tasks)

        # Add interfaces to address family index.
        for name in self.names:
            interface = self.names[name]
            self.add(interface)

        self.started = True
        return self

if __name__ == "__main__": # pragma: no cover
    async def test_interface():
        #out = await cmd("route print")
        out = await nt_route_print("Realtek Gaming 2.5GbE Family Controller")
        print(out)
        return

        i = Interface(AF_ANY)   
        await i.start()
        #af = i.stack if i.stack != DUEL_STACK else IP4
        #b = Bind(i, af)

        return
        if1 = await Interface("enp3s0").start()
        if2 = await Interface("wlp2s0").start()
        ifs = Interfaces([if1, if2])
        print(ifs.by_af)

    async_test(test_interface)




from p2pd.test_init import *
from p2pd import IP6, IP4, Route, Interface
from p2pd.net import AF_ANY, VALID_AFS, VALID_ANY_ADDR
from p2pd.errors import *
from p2pd.nat import nat_info
from p2pd.utils import what_exception
from p2pd.interface import load_interfaces, filter_trash_interfaces
from p2pd.interface import if_list_to_dict, dict_to_if_list

asyncio.set_event_loop_policy(SelectorEventPolicy())
class TestInterface(unittest.IsolatedAsyncioTestCase):
    # Should find at least a valid iface on whatever OS.
    async def test_default_interface(self):
        i = await Interface().start_local()
        self.assertTrue(i.name)

        i = await Interface("").start_local()
        self.assertTrue(i.name)

        i = await Interface(AF_ANY).start_local()
        self.assertTrue(i.name)

        i = await Interface().start_local()
        d_list = if_list_to_dict([i])
        if_list = dict_to_if_list(d_list)

    async def test_invalid_interface_name(self):
        test_passes = False
        try:
            await Interface("meow").start_local()
        except InterfaceNotFound:
            test_passes = True

        self.assertTrue(test_passes)

    async def test_nat_validation(self):
        nat = nat_info()
        i = Interface(nat=nat)
        i.set_nat(nat)

    async def test_fake_ext_route(self):
        one_valid = False
        for af in VALID_AFS:
            try:
                # Throws if no default route for AF.
                i = await Interface(af).start()

                # Throws if no NIC IP for AF.
                r = i.route(af)
                one_valid = True
            except Exception:
                continue

        self.assertTrue(one_valid)

    async def test_win_netifaces_bypass(self):
        pass

    async def test_interface_start(self):
        start_worked = False
        for af in VALID_AFS:
            try:
                await Interface(af).start()
                start_worked = True
                break
            except Exception:
                pass

        self.assertTrue(start_worked)

    async def test_load_interfaces(self):
        ifs = await load_interfaces()
        self.assertTrue(len(ifs))

        # Check nic IP fetch. 
        i = ifs[0]
        af = i.supported()[0]
        n = i.nic(af)
        self.assertTrue(n is not None)

        # Check 'scope id' fetch (used for IPv6 bind code.)
        scope_id = i.get_scope_id()
        self.assertTrue(scope_id is not None)

        # Test loading NAT info for an interface.
        nat = await i.load_nat()
        self.assertTrue(nat is not None)

        # Test dict and load.
        x = eval(repr(i))
        self.assertTrue(isinstance(x, Interface))
        as_s = str(i)
        self.assertTrue(as_s in repr(i))

if __name__ == '__main__':
    main()



from p2pd import *

async def example():
    p = 10233
    d = Daemon()
    i = await Interface().start()
    b = await i.route(i.supported()[0]).bind(ips="127.0.0.1")
    await d.listen_specific(
        targets=[[b, TCP]],
    )
    #
    await d.close()

if __name__ == '__main__':
    async_test(example)


import secrets
import asyncio
from concurrent.futures import ProcessPoolExecutor
import multiprocessing
from decimal import Decimal as Dec
from .p2p_pipe import *
from .daemon import *
from .p2p_protocol import *
from .signaling import *
from .irc_dns import IRCDNS, IRCRefresher, IRC_SERVERS, IRC_REGISTER_FAILURE

NODE_CONF = dict_child({
    # Reusing address can hide errors for the socket state.
    # This can make servers appear to be broken when they're not.
    "reuse_addr": False,
}, NET_CONF)

class P2PUtils():
    # Accomplishes port forwarding and pin hole rules.
    async def forward(self, port):
        tasks = []
        for server in self.servers:
            # Get the bind IP and interface for the route.
            route = server[0]

            # Only forward to public IPv6 addresses.
            ipr = IPRange(route.bind_ip(route.af))
            if route.af == IP6 and ipr.is_private:
                continue

            # Make task to forward this route.
            task = route.forward(port)
            tasks.append(task)

        # Get a list of tasks to do forwarding or pin holes.
        results = await asyncio.gather(*tasks)
        tasks = []
        for result in results:
            if len(result):
                tasks += result

        # Now do that all at once since it might take a while.
        if len(tasks):
            await asyncio.gather(*tasks)

    # Given a dict of credentials find the associated client.
    def find_turn_client(self, turn_server, af=None, interface=None):
        for turn_client in self.turn_clients:
            # Get credentials for turn client.
            credentials = turn_client.get_turn_server(af=af)

            # Compare credentials to turn_server.
            is_same = find_turn_server(credentials, [turn_server])
            if not is_same:
                continue

            # Not the same interface.
            if interface is not None:
                if turn_client.turn_pipe.route.interface != interface:
                    continue

            # Otherwise found it.
            return turn_client

        return None

    def find_signal_pipe(self, addr):
        our_offsets = list(self.signal_pipes)
        for offset in addr["signal"]:
            if offset in our_offsets:
                return self.signal_pipes[offset]

        return None

    def rm_pipe_id(self, pipe_id):
        def closure(data, client_tup, pipe):
            # Delete reference to pipe.
            log(f"Running rm pipe id {pipe_id}")
            if pipe_id in self.pipes:
                del self.pipes[pipe_id]

        return closure

    def setup_multiproc(self, pp_executor, mp_manager):
        # Process pools are disabled.
        if pp_executor is None:
            self.pp_executor = None
            self.mp_manager = None
            return
            
        assert(mp_manager)
        self.pp_executor = pp_executor
        self.mp_manager = mp_manager

    def setup_coordination(self, sys_clock):
        self.sys_clock = sys_clock

    def setup_tcp_punching(self):
        self.tcp_punch_clients = [
            TCPPunch(
                interface,
                self.if_list,
                self.sys_clock,
                self.pp_executor,
                self.mp_manager
            )
            for interface in self.if_list
        ]

# Main class for the P2P node server.
class P2PNode(Daemon, P2PUtils):
    def __init__(self, if_list, port=NODE_PORT, node_id=None, ip=None, signal_offsets=None, enable_upnp=False, conf=NODE_CONF, seed=None):
        super().__init__()
        self.seed = seed or secrets.token_bytes(24)
        self.irc_dns = IRCDNS(if_list[0], self.seed, IRC_SERVERS)
        self.irc_refresher = IRCRefresher(self.irc_dns)
        self.conf = conf
        self.signal_offsets = signal_offsets
        self.port = port
        self.enable_upnp = enable_upnp
        self.ip = ip
        self.if_list = if_list
        self.ifs = Interfaces(if_list)
        self.node_id = node_id or rand_plain(15)
        self.signal_pipes = {} # offset into MQTT_SERVERS
        self.expected_addrs = {} # by [pipe_id]
        self.pipe_events = {} # by [pipe_id]
        self.pipes = {} # by [pipe_id]
        self.sys_clock = None
        self.pp_executor = None
        self.mp_manager = None
        self.tcp_punch_clients = None # [...]
        self.turn_clients = []

        # Handlers for the node's custom protocol functions.
        self.msg_cbs = []

        # Maps p2p's ext to pipe_id.
        self.pending_pipes = {} # by [pipe_id]
        self.is_running = True
        self.listen_all_task = None
        self.signal_worker_task = None
        self.signal_worker_tasks = {} # offset into MQTT_SERVERS

    # Used by the MQTT clients.
    async def signal_protocol(self, msg, signal_pipe):
        return await async_wrap_errors(
            signal_protocol(self, msg, signal_pipe)
        )

    # Used by the node servers.
    async def msg_cb(self, msg, client_tup, pipe):
        return await node_protocol(self, msg, client_tup, pipe)

    # Add custom protocol handlers to node server.
    def add_msg_cb(self, msg_cb):
        self.msg_cbs.append(msg_cb)

    # Del custom protocol handlers from node server.
    def del_msg_cb(self, msg_cb):
        if msg_cb in self.msg_cbs:
            self.msg_cbs.remove(msg_cb)

    # Start the node -- must have been setup properly first.
    async def start(self, protos=[TCP]):
        # MQTT server offsets to try.
        if self.signal_offsets is None:
            offsets = [0] + shuffle([i for i in range(1, len(MQTT_SERVERS))])
        else:
            offsets = self.signal_offsets

        # Get list of N signal pipes.
        for _ in range(0, SIGNAL_PIPE_NO):
            async def set_signal_pipe(offset):
                mqtt_server = MQTT_SERVERS[offset]
                signal_pipe = SignalMock(
                    peer_id=to_s(self.node_id),
                    f_proto=self.signal_protocol,
                    mqtt_server=mqtt_server
                )

                try:
                    await signal_pipe.start()
                    self.signal_pipes[offset] = signal_pipe
                    return signal_pipe
                except Exception:
                    if signal_pipe.is_connected:
                        await signal_pipe.close()
                    return None
            
            # Traverse list of shuffled server indexes.
            while 1:
                # If tried all then we're done.
                if not len(offsets):
                    break

                # Get the next offset to try.
                offset = offsets.pop(0)

                # Try to use the server at the offset for a signal pipe.
                signal_pipe = await async_wrap_errors(
                    set_signal_pipe(offset)
                )

                # The connection failed so keep trying.
                if signal_pipe is None:
                    continue
                else:
                    break

        # Check at least one signal pipe was set.
        if not len(self.signal_pipes):
            raise Exception("Unable to get any signal pipes.")

        # Make a list of routes based on supported address families.
        routes = []
        if_names = []
        for interface in self.if_list:
            for af in interface.supported():
                route = await interface.route(af).bind()
                routes.append(route)

            if_names.append(interface.name)

        # Do deterministic bind to port by NIC IPs.
        if self.port == -1:
            self.port = get_port_by_ips(if_names)
            for _ in range(0, 2):
                try:
                    self.listen_all_task = await self.listen_all(
                        routes,
                        [self.port],
                        protos
                    )
                    break
                except Exception:
                    # Use any port.
                    log(f"Deterministic bind for node server failed.")
                    log(f"Port {self.port} was not available.")
                    self.port = 0
                    log_exception()
        else:
            # Start handling messages for self.msg_cb.
            # Bind to all ifs provided to class on route[0].
            self.listen_all_task = await self.listen_all(
                routes,
                [self.port],
                protos
            )

        # Translate any port 0 to actual assigned port.
        # First server, field 3 == base_proto.
        # sock = listen sock, getsocketname = (bind_ip, bind_port, ...)
        port = self.servers[0][2].sock.getsockname()[1]
        self.addr_bytes = make_peer_addr(self.node_id, self.if_list, list(self.signal_pipes), port=port, ip=self.ip)
        self.p2p_addr = parse_peer_addr(self.addr_bytes)
        log(f"> P2P node = {self.addr_bytes}")

        # Do port forwarding if enabled.
        # Maybe this should go in the route class?
        if self.enable_upnp:
            # UPnP has no unit tests so wrap all errors for now.
            await async_wrap_errors(
                self.forward(port)
            )
            
        return self

    # Connect to a remote P2P node using a number of techniques.
    async def connect(self, addr_bytes, strategies=P2P_STRATEGIES, timeout=60):
        # Assume that its a IRCDNS name.
        if isinstance(addr_bytes, list):
            addr_bytes.append("") # Optional that may have been forgotten.
            ret = await self.irc_dns.name_lookup(*addr_bytes[:3])
            addr_bytes = ret["msg"]

        # Create a TCP connection to the peer using the strategies
        # defined in the strategies list.
        p2p_pipe = P2PPipe(self)
        return await p2p_pipe.pipe(
            addr_bytes,
            strategies=strategies,
            timeout=timeout
        )
    
    """
    Register this name on up to N IRC servers if needed.
    Then set the value at that name to this nodes address
    """
    async def register(self, name_field):
        name_field.append("") # Optional that they may have forgotten.
        out, status = await self.irc_dns.name_register(*name_field[:3])
        if status != IRC_REGISTER_FAILURE:
            name_field.insert(0, self.address)
            await self.irc_dns.store_value(*name_field[:4])
        
        return out, status
    
    async def refresh_names(self):
        await self.irc_refresher.refresher()

    # Get our node server's address.
    def address(self):
        return self.addr_bytes

    # Shutdown the node server and do cleanup.
    async def close(self):
        # Stop node server.
        await super().close()
        self.is_running = False

        # Close the MQTT client.
        for offset in list(self.signal_pipes):
            signal_pipe = self.signal_pipes[offset]
            await signal_pipe.close()

        # Close TCP punch clients.
        for punch_client in self.tcp_punch_clients:
            # Sets close event.
            # Waits for stop event.
            await punch_client.close()

        # Close TURN clients.
        for turn_client in self.turn_clients:
            # Sets state transition to error state to end msg check loop.
            # Closes the open TURN client handle.
            if turn_client is not None:
                await turn_client.close()

        # Cancel all pending p2p_pipes from this node.
        for pipe_task in self.pending_pipes.values():
            pipe_task.cancel()

        """
        Close all open pipes. These may be a mixture of
        inbound cons, outbound cons, servers, or various
        clients. Anything already closed will return.
        """
        for pipe in self.pipes.values():
            await pipe.close()

        # Try close the multiprocess manager.
        try:
            if self.mp_manager is not None:
                self.mp_manager.shutdown()
        except Exception:
            pass

        # Try close the process pool executor.
        try:
            if self.pp_executor is not None:
                self.pp_executor.shutdown()
        except Exception:
            pass

        """
        Node close will throw: 
        Exception ignored in: <function BaseEventLoop.__del__
        with socket error -1

        So you need to make sure to wrap coroutines for exceptions.
        """
        self.mp_manager = None
        self.pp_executor = None
        await asyncio.sleep(.25)

if __name__ == "__main__": # pragma: no cover
    from .p2p_pipe import P2PPipe, P2P_DIRECT, P2P_REVERSE, P2P_PUNCH, P2P_RELAY
    from .nat import *
    async def test_p2p_node():
        sys_clock = SysClock(Dec("-0.02839018452552057081653225806"))
        pp_executor = ProcessPoolExecutor()
        mp_manager = multiprocessing.Manager()
        #internode_out = await Interface("enp3s0").start()
        starlink_out = await Interface("wlp2s0").start()
        internode_nat = nat_info(PRESERV_NAT, delta_info(PRESERV_NAT, 0))
        starlink_nat = nat_info(RESTRICT_PORT_NAT, delta_info(RANDOM_DELTA, 0), [34000, MAX_PORT])
        #internode_out.set_nat(internode_nat)
        starlink_out.set_nat(starlink_nat)





        alice_node = await P2PNode([starlink_out], 11111).start()
        alice_node.setup_multiproc(pp_executor, mp_manager)
        alice_node.setup_coordination(sys_clock)
        alice_node.setup_tcp_punching()

        
        bob_node = await P2PNode([starlink_out], 22222).start()
        bob_node.setup_multiproc(pp_executor, mp_manager)
        bob_node.setup_coordination(sys_clock)
        bob_node.setup_tcp_punching()
        alice_p2p_pipe = P2PPipe(alice_node)
        

        #print(alice_node.tcp_punch_clients[0].interface.route(IP4))



        """
        print("Testing direct connect.")
        direct_pipe = await alice_p2p_pipe.pipe(
            bob_node.addr_bytes,
            [P2P_DIRECT]
        )
        print(direct_pipe)
        """

        """
        print("Testing reverse connect.")
        reverse_pipe = await alice_p2p_pipe.pipe(
            bob_node.addr_bytes,
            [P2P_REVERSE]
        )
        print(reverse_pipe)
        """

        """
        print("Testing p2p punch.")
        punch_pipe = await alice_p2p_pipe.pipe(
            bob_node.addr_bytes,
            [P2P_PUNCH]
        )
        """

        print("Testing p2p relay.")
        relay_pipe = await alice_p2p_pipe.pipe(
            bob_node.addr_bytes,
            [P2P_RELAY]
        )

        print("Got this pipe from relay = ")
        print(relay_pipe)
        
        await relay_pipe.send(b"Test msg back to bob via turn chan.")   
        print("sent")
        while 1:
            await asyncio.sleep(1)

    async_test(test_p2p_node)




from p2pd.test_init import *
from p2pd.nat import *

class TestNAT(unittest.IsolatedAsyncioTestCase):
    async def test_preserving_delta(self):
        fake_stun = FakeSTUNClient(interface=None)
        fake_stun.set_mappings([
            [ 4000, 50000 ],
            [ 10000, 56000 ],
            [ 10200, 56200 ],
            [ 10400, 56400 ],
            [ 11000, 57000 ], # threshold passed.
            [ 12000, 58000 ],
            [ 4620, 2333 ], # wrong
            [ 4630, 2334 ], # wrong
        ])

        expected = delta_info(PRESERV_DELTA, 0)
        got = await delta_test(fake_stun, concurrency=False)
        self.assertEqual(expected, got)

    async def test_equal_delta(self):
        fake_stun = FakeSTUNClient(interface=None)
        fake_stun.set_mappings([
            [ 4567, 4567 ],
            [ 48234, 48234 ],
            [ 6823, 6823 ],
            [ 60000, 60000 ],
            [ 50000, 50000 ], # threshold passed.
            [ 60001, 60001 ],
            [ 5000, 5001 ], # wrong
            [ 5000, 5001 ], # wrong
        ])

        expected = delta_info(EQUAL_DELTA, 0)
        got = await delta_test(fake_stun, concurrency=False)
        self.assertEqual(expected, got)

    async def test_independent_delta(self):
        fake_stun = FakeSTUNClient(interface=None)
        fake_stun.set_mappings([
            [ 5000, 50000 ],
            [ 23412, 50010 ],
            [ 55421, 50020 ],
            [ 6322, 50030 ],
            [ 5622, 50040 ], # threshold passed.
            [ 45610, 50050 ],
            [ 4420, 2333 ], # wrong
            [ 5555, 2334 ], # wrong
        ])

        expected = delta_info(INDEPENDENT_DELTA, 10)
        got = await delta_test(fake_stun, concurrency=False)
        self.assertEqual(expected, got)

    async def test_dependent_delta(self):
        fake_stun = FakeSTUNClient(interface=None)
        fake_stun.set_mappings([
            [ 4560, 50000 ],
            [ 4561, 50020 ],
            [ 4562, 50040 ],
            [ 4563, 50060 ],
            [ 4564, 50080 ], # threshold passed.
            [ 4565, 50100 ],
            [ 4566, 2333 ], # wrong
            [ 4567, 6000 ], # wrong
        ])

        expected = delta_info(DEPENDENT_DELTA, 20)
        got = await delta_test(fake_stun, concurrency=False)
        self.assertEqual(expected, got)

    async def test_random_delta(self):
        fake_stun = FakeSTUNClient(interface=None)
        fake_stun.set_mappings([
            [ 4560, 4443 ],
            [ 4561, 50001 ],
            [ 4562, 63813 ],
            [ 4563, 64000 ],
            [ 4564, 2000 ], # threshold passed.
            [ 4565, 6004 ],
            [ 4566, 2333 ], # wrong
            [ 4567, 8432 ], # wrong
        ])

        expected = delta_info(RANDOM_DELTA, 0)
        got = await delta_test(fake_stun, concurrency=False)
        self.assertEqual(expected, got)

    async def test_nat_intersect_range(self):
        tests = [
            [
                [2000, 10000],
                [5000, 20000],
                [5000, 10000]
            ],
            [
                [2000, 10000],
                [2000, 10000],
                [2000, 10000],
            ],
            [
                [2000, 10000],
                [9999, 10000],
                [9999, 10000],
            ],
        ]

        for test in tests:
            nat_a = nat_info(OPEN_INTERNET,
                delta_info(NA_DELTA, 0),
                test[0]
            )

            nat_b = nat_info(OPEN_INTERNET,
                delta_info(NA_DELTA, 0),
                test[1]
            )

            expected = test[2]
            got = nats_intersect_range(nat_a, nat_b, 0)
            self.assertEqual(got, expected)

    async def test_get_single_mapping(self):
        # Attempt to mimic other sides remote ports.
        mode = TCP_PUNCH_REMOTE

        # What their mappings look like.
        # [ remote, reply, local ] of peers mapping.
        rmap = [ 13073, 0, 50000 ]

        # Last mapped used as starting point for some NAT types.
        # First or last STUN lookup values.
        # [ local, remote ]
        last_mapped = [ 44333, 30000 ]

        # Range that their mappings are in (assume full range.)
        use_range = [ 5000, 68000 ]

        # Fake STUN client used to test logic paths.
        i = None; af = IP4; our_nat = None;
        stun_client = FakeSTUNClient(interface=i, af=af)

        # All params to pass to func.
        params = [ mode, rmap, last_mapped, use_range, our_nat, stun_client ]
        f_check = lambda e, r: r[0][:3] == e

        ##########################################################
        # Case 1 = Our NAT is fully open.
        params[-2] = nat_info(OPEN_INTERNET, delta_info(NA_DELTA, 0))
        # [ local, remote, reply ]
        expected = [ rmap[0], rmap[0], 0 ]
        ret = await get_single_mapping(*params)
        self.assertTrue(f_check(expected, ret))

        ##########################################################
        # Case 2 = Our NAT can preserve local ports.
        params[-2] = nat_info(FULL_CONE, delta_info(EQUAL_DELTA, 0))
        expected = [ rmap[0], rmap[0], 0 ]
        ret = await get_single_mapping(*params)
        self.assertTrue(f_check(expected, ret))

        ##########################################################
        # Case 3 = Our delta preserves the distance between locals.
        params[-2] = nat_info(RESTRICT_PORT_NAT, delta_info(PRESERV_DELTA, 0))
        """
        n_dist = ( last_remote (30000), bind_port (13073) )
               = 16927
        last_local = 44333 (maps to the last remote)
        next_local = port_wrap(last_local + n_dist)
                   = 61260
        """
        expected = [ 61260, rmap[0], rmap[0] ]
        ret = await get_single_mapping(*params)
        self.assertTrue(f_check(expected, ret))

        ##########################################################
        # Case 4 = Each new mapping is + delta value.
        delta_val = 4
        params[-2] = nat_info(
            RESTRICT_NAT,
            delta_info(INDEPENDENT_DELTA, delta_val)
        )
        ret = await get_single_mapping(*params)
        expected = [ 
            ret[0][0],
            last_mapped[1] + delta_val,
            0
        ]
        self.assertTrue(f_check(expected, ret))

        ##########################################################
        # Case 5 = Each new mapping is + delta value.
        # Only if local ports also increase sequentially.
        params[-2] = nat_info(
            RESTRICT_NAT,
            delta_info(DEPENDENT_DELTA, delta_val)
        )
        expected = [ 
            last_mapped[0] + 1,
            last_mapped[1] + delta_val,
            0
        ]
        ret = await get_single_mapping(*params)
        self.assertTrue(f_check(expected, ret))

        ##########################################################
        # Case 6 = Reusable mapping from same local port.
        params[-2] = nat_info(
            FULL_CONE,
            delta_info(RANDOM_DELTA, 0)
        )

        stun_client.set_mappings([[44444, 33333]])
        expected = [44444, 33333, 0 ]
        ret = await get_single_mapping(*params)
        self.assertTrue(f_check(expected, ret))

        

    
if __name__ == '__main__':
    main()

"""
is_valid_rmap

type 5 nat and random delta inside a vmware 'nat' vm behind a router with a cone type nat and preserving delta.
"""



import asyncio
from concurrent.futures import ProcessPoolExecutor
from .p2p_node import *

def init_process_pool():
    # Make selector default event loop.
    # On Windows this changes it from proactor to selector.
    asyncio.set_event_loop_policy(SelectorEventPolicy())

    # Create new event loop in the process.
    loop = asyncio.get_event_loop()

    # Handle exceptions on close.
    loop.set_exception_handler(handle_exceptions)

async def get_pp_executors(workers=2):
    try:
        pp_executor = ProcessPoolExecutor(max_workers=workers)
    except Exception:
        """
        Not all platform have a working implementation of sem_open / semaphores.
        Android is one such platform. It does support multiprocessing but
        this semaphore feature is missing and will throw an error here.
        In this case -- log the error and revert to using a single event loop.
        """
        log_exception()
        return None
    
    loop = asyncio.get_event_loop()
    tasks = []
    for i in range(0, workers):
        tasks.append(loop.run_in_executor(
            pp_executor, init_process_pool
        ))
    await asyncio.gather(*tasks)
    return pp_executor

# delay with sys clock and get_pp_executors.
async def start_p2p_node(port=NODE_PORT, node_id=None, ifs=None, clock_skew=Dec(0), ip=None, pp_executors=False, enable_upnp=False, signal_offsets=None, netifaces=None):
    netifaces = netifaces or Interface.get_netifaces()
    if netifaces is None:
        netifaces = await init_p2pd()

    # Load NAT info for interface.
    ifs = ifs or await load_interfaces(netifaces=netifaces)
    assert(len(ifs))
    for interface in ifs:
        # Don't set NAT details if already set.
        if interface.resolved:
            continue

        # Prefer IP4 if available.
        af = IP4
        if af not in interface.supported():
            af = IP6

        # STUN is used to test the NAT.
        stun_client = STUNClient(
            interface,
            af
        )

        # Load NAT type and delta info.
        # On a server should be open.
        nat = await stun_client.get_nat_info()
        interface.set_nat(nat)

    if pp_executors is None:
        pp_executors = await get_pp_executors(workers=4)

    if clock_skew == Dec(0):
        sys_clock = await SysClock(ifs[0]).start()
    else:
        sys_clock = SysClock(ifs[0], clock_skew)

    # Log sys clock details.
    assert(sys_clock.clock_skew) # Must be set for meetings!
    log(f"> Start node. Clock skew = {sys_clock.clock_skew}")

    # Pass interface list to node.
    node = await async_wrap_errors(
        P2PNode(
            if_list=ifs,
            port=port,
            node_id=node_id,
            ip=ip,
            signal_offsets=signal_offsets,
            enable_upnp=enable_upnp
        ).start()
    )

    log("node success apparently.")

    # Configure node for TCP punching.
    if pp_executors is not None:
        mp_manager = multiprocessing.Manager()
    else:
        mp_manager = None

    node.setup_multiproc(pp_executors, mp_manager)
    node.setup_coordination(sys_clock)
    node.setup_tcp_punching()

    # Wait for MQTT sub.
    for offset in list(node.signal_pipes):
        await node.signal_pipes[offset].sub_ready.wait()

    return node




"""
When using UDP in network programming its common to write
code so that it has simple loops to retry sending packets if
no response is received. The problem is: if any packets are lost
the time spent waiting keeps accumulating to the cost of a round-trip.

The problem is made worse when you consider slow-downs in DNS and
the often unreliable nature of relying on community servers for
STUN code. If one is not careful they quickly end up with an
algorithm that is prohibitively slow in the best case and quite
unreliable in the worst case.

The original algorithm for doing STUN tests for RFC 3489 used
incremental tests and was very brittle. However -- due to the
nature of how these tests work -- it became clear to me that it was
possible to paralyze the tests into two main phases if a STUN
server's primary and secondary IP were known beforehand.

Phase 1 tests for [open NAT and full cone] while phase 2 tests for
[restrict ip and restrict port] behaviors. The algorithm here is designed to run across multiple public servers where it creates races among
all servers to test a routes NAT. The result is you end up with
the fastest possible determination of NAT behaviors while also
building in safe-guards against packet-loss, inconsistent results, misconfigurations, and slow network conditions.
"""

import time
from .settings import *
from .ip_range import *
from .stun_client import *

# Constants for a NAT test.
NAT_TEST_NO = 5
NAT_TEST_TIMEOUT = 0.5

# STUN payload to send, Send IP, send port, reply IP, reply port
# Shows order of RFC 3489 NAT enumeration sets.
NAT_TEST_SCHEMA = [
    # Detects: open NAT.
    ["", "primary", "primary", "primary", "primary"],

    # Detects: full cone NAT.
    # Change both reply IP and port.
    [changeRequest, "primary", "primary", "secondary", "secondary"],

    # Detects: non-symmetric NAT.
    ["", "secondary", "primary", "secondary", "primary"],

    # Detects: between restrict and port restrict.
    # Change only the reply port.
    [changePortRequest, "secondary", "primary", "secondary", "secondary"],
]

def filter_nat_servers(nat_servers):
    out = []

    def find_duplicates(hey, cmp_ipr=None):
        ip_types = ["primary", "secondary"]
        for entry in hey:
            for ip_type in ip_types:
                entry_ip = entry[ip_type]["ip"]
                if entry_ip is None:
                    continue
                else:
                    entry_ipr = IPRange(entry_ip)

                if cmp_ipr is not None:
                    if entry_ipr == cmp_ipr:
                        return True
                else:
                    if not find_duplicates(out, entry_ipr):
                        out.append(entry)

        return False
            
    find_duplicates(nat_servers)
    return out

async def new_stun_server_format(af):
    servers = []
    for index in range(0, len(STUND_SERVERS[af])):
        route = i.route(af)
        stun_addr = await Address(
            STUND_SERVERS[af][index][0],
            STUND_SERVERS[af][index][1],
            route
        )

        out = await do_nat_test(
            stun_addr=stun_addr,
            interface=i,
            af=af,
            proto=TCP,
            group="change"
        )

        print(out)

        if out is None or isinstance(out, tuple):
            continue
    
        svr = {
            "host": STUND_SERVERS[af][index][0],
            "primary": {
                "ip": stun_addr.tup[0],
                "port": stun_addr.tup[1]
            },
            "secondary": {
                "ip": out["cip"],
                "port": out["cport"]
            }
        }

        servers.append(svr)


    print(servers)

async def nat_test_exec(dest_addr, reply_addr, payload, pipe, q, test_coro):
    # Expect messages from this reply_addr.
    tran_info = tran_info_patterns(reply_addr.tup)
    pipe.subscribe(tran_info[0:2])
    conf = dict_merge(STUN_CONF, {
        "packet_retry": 1,
        "recv_timeout": NAT_TEST_TIMEOUT
    })

    # Do first NAT test.
    ret, _ = await stun_sub_test(
        "running nat test",
        dest_addr,
        pipe.route.interface,
        pipe.route.af,
        UDP,
        pipe.sock.getsockname()[1],
        reply_addr,
        payload,
        pipe=pipe,
        tran_info=tran_info,
        conf=conf
    )

    # Valid reply.
    if ret is not None and not isinstance(ret, tuple):
        if ret["resp"]:
            q.append(ret)
            return await test_coro(ret, pipe)

    return None

async def nat_test_workers(pipe, q, test_index, test_coro, servers):
    # Make list of coroutines to do this NAT tests.
    workers = []
    for server_no in range(0, min(NAT_TEST_NO, len(servers))):
        async def worker(server_no):
            # Packets will go to this destination.
            # Send to, expect from.
            addrs = [] 
            for x in range(0, 2):
                # Determine which fields to use for IP and port.
                schema = NAT_TEST_SCHEMA[test_index][(x * 2) + 1:(x * 2) + 3]
                ip_type = schema[0]
                port_type = schema[1]

                # Resolve IP and port as an Address.
                addrs.append(
                    await Address(
                        servers[server_no][ip_type]["ip"],
                        servers[server_no][port_type]["port"],
                        pipe.route
                    )
                )

            # Run the test and return the results.
            payload = NAT_TEST_SCHEMA[test_index][0]
            return await nat_test_exec(
                # Send to and expect from.
                addrs[0],
                addrs[1],

                # Type of STUN request to send.
                payload,

                # Pipe to reuse for UDP.
                pipe,

                # Async queue to store the results.
                q,

                # Test-specific code.
                test_coro
            )

        workers.append(worker(server_no))

    return workers

"""
If a NAT uses the same 'mapping' (external IP and port) given
the same internal (IP and port) even when destinations are
different then it's considered non-symmetric. The software
proceeds to determine the exact conditions for which mappings
can be reused when using the same bind tuples.
"""
def non_symmetric_check(q_list):
    # Test 1 and 3.
    q1 = q_list[0]
    q3 = q_list[2]

    # Not enough data to know.
    if not len(q1) or not len(q3):
        return False

    # NAT reuses mappings given same internal (ip and port)
    port_check = q1[0]['rport'] == q3[0]['rport']
    ip_check = ip_f(q1[0]['rip']) == ip_f(q3[0]['rip'])
    if port_check and ip_check:
        return True

    # Otherwise return False.
    return False

"""
If there's no replies in any of the NAT test lists then
assume that this means there's a firewall and return False.
"""
def no_stun_resp_check(q_list):
    for i in range(0, 4):
        if len(q_list[i]):
            return False
        
    return True

async def fast_nat_test(pipe, test_servers, timeout=NAT_TEST_TIMEOUT):
    # Ensure all primary and secondary IPs are unique
    test_servers = filter_nat_servers(test_servers)

    # Store STUN request results here.
    # n = index of test e.g. [0] = test 1.
    q_list = [[], [], [], []]
    q_list.append(SYMMETRIC_NAT)

    # Open NAT type.
    async def test_one(ret, test_pipe):
        source_ip = test_pipe.route.bind_ip()
        if ip_f(ret['rip']) == ip_f(source_ip):
            return OPEN_INTERNET

    # Full cone NAT.
    async def test_two(ret, test_pipe):
        # Test 2 may arrive before test 1.
        # In this case: test 1 takes priority over test 2.
        return await test_one(ret, test_pipe) or FULL_CONE
    
    # Whitelist of dest (IP and port).
    async def test_three(ret, test_pipe):
        if non_symmetric_check(q_list):
            q_list[4] = RESTRICT_PORT_NAT

    # Whitelist of dest (IP).
    async def test_four(ret, test_pipe):
        return RESTRICT_NAT
    
    """
    All tests in sub_test_a are tried then sub_tests_b.
    Both sub test lists can't be run concurrently
    due to how NATs function with white listing.
    """
    test_index = 0
    for sub_test in [[test_one, test_two], [test_three, test_four]]:
        # Get a list of workers for first two NAT tests.
        workers = []
        for test_coro in sub_test:
            # Build list of coroutines to run these NAT tests.
            # Test funcs are run on receiving a STUN response.
            workers += await nat_test_workers(
                pipe,
                q_list[test_index],
                test_index,
                test_coro,
                test_servers, 
            )

            # Keep track of test offset.
            test_index += 1

        # Run NAT sub tests.
        try:
            # First result in or timeout.
            for task in asyncio.as_completed(workers, timeout=timeout):
                ret = await task
                if ret is not None:
                    return ret
        except asyncio.TimeoutError:
            continue

    # All tests timed out.
    # Determine return value.
    if no_stun_resp_check(q_list):
        return BLOCKED_NAT
    else:
        # Symmetric NAT or RESTRICT_PORT_NAT.
        return q_list[-1] 

async def nat_test_main():
    from .interface import Interface, init_p2pd

    # Load internal interface details.
    t1 = timestamp(1)
    netifaces = await init_p2pd()
    t2 = timestamp(1)
    duration = t2 - t1
    print(f"init_p2pd() = {duration}")

    # Start interface time.
    i = await Interface("ens33", netifaces=netifaces) # ens37
    t1 = timestamp(1)
    nat = await i.load_nat()
    print(nat)
    t2 = timestamp(1)
    duration = t2 - t1
    print(f"Interface() load_nat = {duration}")
    await asyncio.sleep(NAT_TEST_NO)
    return


    # Use same pipe with multiplexing for reuse tests.
    #t1 = timestamp(1)
    af = IP4
    route = await i.route(af).bind(0)
    pipe = await pipe_open(UDP, route=route, conf=STUN_CONF)
    assert(pipe is not None)

    # Determine NAT type.
    t1 = timestamp(1)
    servers = STUND_SERVERS[af]
    nat_type = await fast_nat_test(pipe, servers)
    print(nat_type)

    # How long the NAT type took.
    t2 = timestamp(1) - t1
    print(f"nat time = {t2}")
    await pipe.close()
    # 0.8

    # Wait for all lagging tests to end.
    await asyncio.sleep(NAT_TEST_TIMEOUT)

if __name__ == "__main__":
    async_test(nat_test_main)